"""Implementation of the equations used for surface tension modelling, for
example in KHI simulations. The references are [SY11], [JM00], [A10], [XA06].


References
----------

.. [SY11] M. Shadloo, M. Yildiz, "Numerical modelling of Kelvin-Helmholtz
   instability using smoothed particle hydrodynamics", IJNME, 2011, 87, pp
   988--1006

.. [JM00] Joseph P. Morris "Simulating surface tension with smoothed particle
   hydrodynamics", JCP, 2000, 33, pp 333--353

.. [A10] Adami et al. "A new surface-tension formulation for multi-phase SPH
   using a reproducing divergence approximation", JCP 2010, 229, pp 5011--5021

.. [XA06] X.Y.Hu, N.A. Adams. "A multi-phase SPH method for macroscopic and
   mesoscopic flows", JCP 2006, 213, pp 844-861 [XA06]

"""

from math import sqrt

from pysph.sph.equation import Group, Equation

from pysph.sph.gas_dynamics.basic import ScaleSmoothingLength

from pysph.sph.wc.transport_velocity import SummationDensity, \
    MomentumEquationPressureGradient, StateEquation,\
    MomentumEquationArtificialStress, MomentumEquationViscosity, \
    SolidWallNoSlipBC

from pysph.sph.wc.linalg import gj_solve, augmented_matrix

from pysph.sph.basic_equations import IsothermalEOS

from pysph.sph.wc.basic import TaitEOS


class SurfaceForceAdami(Equation):
    def initialize(self, d_au, d_av, d_idx):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0

    def loop(self, d_au, d_av, d_aw, d_idx, d_m, DWIJ, d_pi00, d_pi01, d_pi02,
             d_pi10, d_pi11, d_pi12, d_pi20, d_pi21, d_pi22, s_pi00, s_pi01,
             s_pi02, s_pi10, s_pi11, s_pi12, s_pi20, s_pi21, s_pi22, d_V, s_V,
             s_idx):
        s2 = s_V[s_idx]*s_V[s_idx]
        f00 = (d_pi00[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi00[s_idx]/s2)
        f01 = (d_pi01[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi01[s_idx]/s2)
        f02 = (d_pi02[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi02[s_idx]/s2)
        f10 = (d_pi10[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi10[s_idx]/s2)
        f11 = (d_pi11[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi11[s_idx]/s2)
        f12 = (d_pi12[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi12[s_idx]/s2)
        f20 = (d_pi20[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi20[s_idx]/s2)
        f21 = (d_pi21[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi21[s_idx]/s2)
        f22 = (d_pi22[d_idx]/(d_V[d_idx]*d_V[d_idx]) + s_pi22[s_idx]/s2)
        d_au[d_idx] += (DWIJ[0]*f00 + DWIJ[1]*f10 + DWIJ[2]*f20)/d_m[d_idx]
        d_av[d_idx] += (DWIJ[0]*f01 + DWIJ[1]*f11 + DWIJ[2]*f21)/d_m[d_idx]
        d_aw[d_idx] += (DWIJ[0]*f02 + DWIJ[1]*f12 + DWIJ[2]*f22)/d_m[d_idx]


class ConstructStressMatrix(Equation):

    def __init__(self, dest, sources, sigma, d=2):
        self.sigma = sigma
        self.d = d
        super(ConstructStressMatrix, self).__init__(dest, sources)

    def initialize(self, d_pi00, d_pi01, d_pi02, d_pi10, d_pi11, d_pi12,
                   d_pi20, d_pi21, d_pi22, d_cx, d_cy, d_cz, d_idx, d_N):
        mod_gradc2 = d_cx[d_idx]*d_cx[d_idx] + d_cy[d_idx]*d_cy[d_idx] + \
            d_cz[d_idx]*d_cz[d_idx]
        mod_gradc = sqrt(mod_gradc2)
        d_N[d_idx] = 0.0
        if mod_gradc > 1e-14:
            factor = self.sigma/mod_gradc
            d_pi00[d_idx] = (-d_cx[d_idx]*d_cx[d_idx] +
                             (mod_gradc2)/self.d)*factor
            d_pi01[d_idx] = -factor*d_cx[d_idx]*d_cy[d_idx]
            d_pi02[d_idx] = -factor*d_cx[d_idx]*d_cz[d_idx]
            d_pi10[d_idx] = -factor*d_cx[d_idx]*d_cy[d_idx]
            d_pi11[d_idx] = (-d_cy[d_idx]*d_cy[d_idx] +
                             (mod_gradc2)/self.d)*factor
            d_pi12[d_idx] = -factor*d_cy[d_idx]*d_cz[d_idx]
            d_pi20[d_idx] = -factor*d_cx[d_idx]*d_cz[d_idx]
            d_pi21[d_idx] = -factor*d_cy[d_idx]*d_cz[d_idx]
            d_pi22[d_idx] = (-d_cz[d_idx]*d_cz[d_idx] +
                             (mod_gradc2)/self.d)*factor
            d_N[d_idx] = 1.0
        else:
            d_pi00[d_idx] = 0.0
            d_pi01[d_idx] = 0.0
            d_pi02[d_idx] = 0.0
            d_pi10[d_idx] = 0.0
            d_pi11[d_idx] = 0.0
            d_pi12[d_idx] = 0.0
            d_pi20[d_idx] = 0.0
            d_pi21[d_idx] = 0.0
            d_pi22[d_idx] = 0.0


class ColorGradientAdami(Equation):

    def initialize(self, d_idx, d_cx, d_cy, d_cz):
        d_cx[d_idx] = 0.0
        d_cy[d_idx] = 0.0
        d_cz[d_idx] = 0.0

    def loop(self, d_idx, d_cx, d_cy, d_cz, d_V, s_V, d_color, s_color, DWIJ,
             s_idx):
        c_i = d_color[d_idx]/(d_V[d_idx]*d_V[d_idx])
        c_j = s_color[s_idx]/(s_V[s_idx]*s_V[s_idx])
        factor = d_V[d_idx]*(c_i + c_j)
        d_cx[d_idx] += factor*DWIJ[0]
        d_cy[d_idx] += factor*DWIJ[1]
        d_cz[d_idx] += factor*DWIJ[2]


class MomentumEquationViscosityAdami(Equation):

    def initialize(self, d_au, d_av, d_aw, d_idx):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_V, d_au, d_av, d_aw, s_V, d_p, s_p, DWIJ, s_idx,
             d_m, R2IJ, XIJ, EPS, VIJ, d_nu, s_nu):
        factor = 2.0*d_nu[d_idx]*s_nu[s_idx]/(d_nu[d_idx] + s_nu[s_idx])
        V_i = 1/(d_V[d_idx]*d_V[d_idx])
        V_j = 1/(s_V[s_idx]*s_V[s_idx])
        dwijdotrij = (DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2])
        dwijdotrij /= (R2IJ + EPS)
        factor = factor*(V_i + V_j)*dwijdotrij/d_m[d_idx]
        d_au[d_idx] += factor*VIJ[0]
        d_av[d_idx] += factor*VIJ[1]
        d_aw[d_idx] += factor*VIJ[2]


class MomentumEquationPressureGradientHuAdams(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(MomentumEquationPressureGradientHuAdams,
              self).__init__(dest, sources)


    def initialize(self, d_au, d_av, d_aw, d_idx):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_V, d_au, d_av, d_aw, s_V, d_p, s_p, DWIJ, s_idx,
             d_m):
        p_i = d_p[d_idx]/(d_V[d_idx]*d_V[d_idx])
        p_j = s_p[s_idx]/(s_V[s_idx]*s_V[s_idx])
        d_au[d_idx] += -(p_i+p_j)*DWIJ[0]/d_m[d_idx]
        d_av[d_idx] += -(p_i+p_j)*DWIJ[1]/d_m[d_idx]
        d_aw[d_idx] += -(p_i+p_j)*DWIJ[2]/d_m[d_idx]

    def post_loop(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] += self.gx
        d_av[d_idx] += self.gy
        d_aw[d_idx] += self.gz


class MomentumEquationPressureGradientAdami(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(MomentumEquationPressureGradientAdami,
              self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho,
             d_au, d_av, d_aw, d_p, s_p,
             d_V, s_V, DWIJ):

        # averaged pressure Eq. (7)
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        p_i = d_p[d_idx]
        p_j = s_p[s_idx]

        pij = rhoj * p_i + rhoi * p_j
        pij /= (rhoj + rhoi)

        # particle volumes; d_V is inverse volume
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # inverse mass of destination particle
        mi1 = 1.0/d_m[d_idx]

        # accelerations 1st term in Eq. (8)
        tmp = -pij * mi1 * (Vi2 + Vj2)

        d_au[d_idx] += tmp * DWIJ[0]
        d_av[d_idx] += tmp * DWIJ[1]
        d_aw[d_idx] += tmp * DWIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] += self.gx
        d_av[d_idx] += self.gy
        d_aw[d_idx] += self.gz


class MomentumEquationViscosityMorris(Equation):

    def __init__(self, dest, sources, eta=0.01):
        self.eta = eta*eta
        super(MomentumEquationViscosityMorris, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_au, d_av, d_aw, s_m, d_nu, s_nu, d_rho,
             s_rho, DWIJ, R2IJ, VIJ, HIJ, XIJ):
        r2 = R2IJ + self.eta*HIJ*HIJ
        dw = (DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2])/(r2)
        mult = s_m[s_idx]*(d_nu[d_idx] + s_nu[s_idx]) / \
            (d_rho[d_idx]*s_rho[s_idx])
        d_au[d_idx] += dw*mult*VIJ[0]
        d_av[d_idx] += dw*mult*VIJ[1]
        d_aw[d_idx] += dw*mult*VIJ[2]


class MomentumEquationPressureGradientMorris(Equation):

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_au, d_av, d_aw, s_m, d_p, s_p, DWIJ, d_rho,
             s_rho):
        factor = -s_m[s_idx]*(d_p[d_idx] + s_p[s_idx]) / \
            (d_rho[d_idx]*s_rho[s_idx])
        d_au[d_idx] += factor*DWIJ[0]
        d_av[d_idx] += factor*DWIJ[1]
        d_aw[d_idx] += factor*DWIJ[2]


class InterfaceCurvatureFromDensity(Equation):

    def __init__(self, dest, sources, with_morris_correction=True):
        self.with_morris_correction = with_morris_correction

        super(InterfaceCurvatureFromDensity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_kappa, d_wij_sum):
        d_kappa[d_idx] = 0.0
        d_wij_sum[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_kappa, d_nx, d_ny, d_nz, s_nx, s_ny, s_nz,
             d_V, s_V, d_N, s_N, d_wij_sum, s_rho, s_m, WIJ, DWIJ):

        nijdotdwij = (d_nx[d_idx] - s_nx[s_idx]) * DWIJ[0] + \
            (d_ny[d_idx] - s_ny[s_idx]) * DWIJ[1] + \
            (d_nz[d_idx] - s_nz[s_idx]) * DWIJ[2]

        tmp = 1.0
        if self.with_morris_correction:
            tmp = min(d_N[d_idx], s_N[s_idx])

        d_wij_sum[d_idx] += tmp * s_m[s_idx]/s_rho[s_idx] * WIJ

        d_kappa[d_idx] += tmp*nijdotdwij*s_m[s_idx]/s_rho[s_idx]

    def post_loop(self, d_idx, d_wij_sum, d_nx, d_kappa):

        if self.with_morris_correction:
            if d_wij_sum[d_idx] > 1e-12:
                d_kappa[d_idx] /= d_wij_sum[d_idx]


class SolidWallPressureBCnoDensity(Equation):

    def initialize(self, d_idx, d_p, d_wij):
        d_p[d_idx] = 0.0
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_p, s_p, d_wij, s_rho,
             d_au, d_av, d_aw, WIJ, XIJ):

        d_p[d_idx] += s_p[s_idx]*WIJ

        d_wij[d_idx] += WIJ

    def post_loop(self, d_idx, d_wij, d_p, d_rho):
        if d_wij[d_idx] > 1e-14:
            d_p[d_idx] /= d_wij[d_idx]


class SummationDensitySourceMass(Equation):

    def initialize(self, d_idx, d_V, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_V, d_rho, d_m, WIJ, s_m, s_idx):
        d_rho[d_idx] += d_m[d_idx]*WIJ

    def post_loop(self, d_idx, d_V, d_rho, d_m):
        d_V[d_idx] = d_rho[d_idx]/d_m[d_idx]


class SmoothedColor(Equation):
    r"""Smoothed color function. Eq. (17) in [JM00]

    .. math::

        c_a = \sum_b \frac{m_b}{\rho_b} c_b^i \nabla_a W_{ab}\,,

    where, :math:`c_b^i` is the color index associated with a
    particle.

    """

    def initialize(self, d_idx, d_scolor):
        d_scolor[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, s_rho, s_m,
             s_color, d_scolor, WIJ):

        # Smoothed color Eq. (17) in [JM00]
        d_scolor[d_idx] += s_m[s_idx]/s_rho[s_idx] * s_color[s_idx] * WIJ


class ColorGradientUsingNumberDensity(Equation):
    r"""Gradient of the color function using Eq. (13) of [SY11]:

    .. math::

        \nabla C_a = \sum_b \frac{2 C_b - C_a}{\psi_a + \psi_a}
        \nabla_{a} W_{ab}


    Using the gradient of the color function, the normal and
    discretized dirac delta is calculated in the post
    loop.

    Singularities are avoided as per the recommendation by [JM00] (see eqs
    20 & 21) using the parameter :math:`\epsilon`

    """

    def __init__(self, dest, sources, epsilon=1e-6):
        self.epsilon2 = epsilon*epsilon
        super(ColorGradientUsingNumberDensity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_cx, d_cy, d_cz, d_nx, d_ny, d_nz,
                   d_ddelta, d_N):

        # color gradient
        d_cx[d_idx] = 0.0
        d_cy[d_idx] = 0.0
        d_cz[d_idx] = 0.0

        # interface normals
        d_nx[d_idx] = 0.0
        d_ny[d_idx] = 0.0
        d_nz[d_idx] = 0.0

        # discretized dirac delta
        d_ddelta[d_idx] = 0.0

        # reliability indicator for normals
        d_N[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_scolor, s_scolor, d_cx, d_cy, d_cz,
             d_V, s_V, DWIJ):

        # average particle volume
        psiab1 = 2.0/(d_V[d_idx] + s_V[s_idx])

        # difference in color divided by psiab. Eq. (13) in [SY11]
        Cba = (s_scolor[s_idx] - d_scolor[d_idx]) * psiab1

        # color gradient
        d_cx[d_idx] += Cba * DWIJ[0]
        d_cy[d_idx] += Cba * DWIJ[1]
        d_cz[d_idx] += Cba * DWIJ[2]

    def post_loop(self, d_idx, d_cx, d_cy, d_cz,
                  d_nx, d_ny, d_nz, d_N, d_ddelta):
        # absolute value of the color gradient
        mod_gradc2 = d_cx[d_idx]*d_cx[d_idx] + \
            d_cy[d_idx]*d_cy[d_idx] + \
            d_cz[d_idx]*d_cz[d_idx]

        # avoid sqrt computations on non-interface particles
        # (particles for which the color gradient is zero) Eq. (19,
        # 20) in [JM00]
        if mod_gradc2 > self.epsilon2:
            # this normal is reliable in the sense of [JM00]
            d_N[d_idx] = 1.0

            # compute the normals
            mod_gradc = 1./sqrt(mod_gradc2)

            d_nx[d_idx] = d_cx[d_idx] * mod_gradc
            d_ny[d_idx] = d_cy[d_idx] * mod_gradc
            d_nz[d_idx] = d_cz[d_idx] * mod_gradc

            # discretized Dirac Delta function
            d_ddelta[d_idx] = 1./mod_gradc


class MorrisColorGradient(Equation):
    r"""Gradient of the color function using Eq. (17) of [JM00]:

    .. math::

        \nabla c_a = \sum_b \frac{m_b}{\rho_b}(c_b - c_a) \nabla_{a}
        W_{ab}\,,

    where a smoothed representation of the color is used in the
    equation. Using the gradient of the color function, the normal and
    discretized dirac delta is calculated in the post loop.

    Singularities are avoided as per the recommendation by [JM00] (see eqs
    20 & 21) using the parameter :math:`\epsilon`

    """

    def __init__(self, dest, sources, epsilon=1e-6):
        self.epsilon2 = epsilon*epsilon
        super(MorrisColorGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_cx, d_cy, d_cz, d_nx, d_ny, d_nz,
                   d_ddelta, d_N):

        # color gradient
        d_cx[d_idx] = 0.0
        d_cy[d_idx] = 0.0
        d_cz[d_idx] = 0.0

        # interface normals
        d_nx[d_idx] = 0.0
        d_ny[d_idx] = 0.0
        d_nz[d_idx] = 0.0

        # reliability indicator for normals and dirac delta
        d_N[d_idx] = 0.0
        d_ddelta[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_scolor, s_scolor, d_cx, d_cy, d_cz,
             s_m, s_rho, DWIJ):

        # Eq. (17) in [JM00]
        Cba = (s_scolor[s_idx] - d_scolor[d_idx]) * s_m[s_idx]/s_rho[s_idx]

        # color gradient
        d_cx[d_idx] += Cba * DWIJ[0]
        d_cy[d_idx] += Cba * DWIJ[1]
        d_cz[d_idx] += Cba * DWIJ[2]

    def post_loop(self, d_idx, d_cx, d_cy, d_cz,
                  d_nx, d_ny, d_nz, d_N, d_ddelta):
        # absolute value of the color gradient
        mod_gradc2 = d_cx[d_idx]*d_cx[d_idx] + \
            d_cy[d_idx]*d_cy[d_idx] + \
            d_cz[d_idx]*d_cz[d_idx]

        # avoid sqrt computations on non-interface particles
        # (particles for which the color gradient is zero) Eq. (19,
        # 20) in [JM00]
        if mod_gradc2 > self.epsilon2:
            # this normal is reliable in the sense of [JM00]
            d_N[d_idx] = 1.0

            # compute the normals
            mod_gradc = 1./sqrt(mod_gradc2)

            d_nx[d_idx] = d_cx[d_idx] * mod_gradc
            d_ny[d_idx] = d_cy[d_idx] * mod_gradc
            d_nz[d_idx] = d_cz[d_idx] * mod_gradc

            # discretized Dirac Delta function
            d_ddelta[d_idx] = 1./mod_gradc


class SY11ColorGradient(Equation):
    r"""Gradient of the color function using Eq. (13) of [SY11]:

    .. math::

        \nabla C_a = \sum_b \frac{2 C_b - C_a}{\psi_a + \psi_a}
        \nabla_{a} W_{ab}


    Using the gradient of the color function, the normal and
    discretized dirac delta is calculated in the post
    loop.

    """

    def __init__(self, dest, sources, epsilon=1e-6):
        self.epsilon2 = epsilon*epsilon
        super(SY11ColorGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_cx, d_cy, d_cz, d_nx, d_ny, d_nz,
                   d_ddelta, d_N):

        # color gradient
        d_cx[d_idx] = 0.0
        d_cy[d_idx] = 0.0
        d_cz[d_idx] = 0.0

        # interface normals
        d_nx[d_idx] = 0.0
        d_ny[d_idx] = 0.0
        d_nz[d_idx] = 0.0

        # discretized dirac delta
        d_ddelta[d_idx] = 0.0

        # reliability indicator for normals
        d_N[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_color, s_color, d_cx, d_cy, d_cz,
             d_V, s_V, DWIJ):

        # average particle volume
        psiab1 = 2.0/(d_V[d_idx] + s_V[s_idx])

        # difference in color divided by psiab. Eq. (13) in [SY11]
        Cba = (s_color[s_idx] - d_color[d_idx]) * psiab1

        # color gradient
        d_cx[d_idx] += Cba * DWIJ[0]
        d_cy[d_idx] += Cba * DWIJ[1]
        d_cz[d_idx] += Cba * DWIJ[2]

    def post_loop(self, d_idx, d_cx, d_cy, d_cz,
                  d_nx, d_ny, d_nz, d_N, d_ddelta):
        # absolute value of the color gradient
        mod_gradc2 = d_cx[d_idx]*d_cx[d_idx] + \
            d_cy[d_idx]*d_cy[d_idx] + \
            d_cz[d_idx]*d_cz[d_idx]

        # avoid sqrt computations on non-interface particles
        if mod_gradc2 > self.epsilon2:
            # this normal is reliable in the sense of [JM00]
            d_N[d_idx] = 1.0

            # compute the normals
            mod_gradc = 1./sqrt(mod_gradc2)

            d_nx[d_idx] = d_cx[d_idx] * mod_gradc
            d_ny[d_idx] = d_cy[d_idx] * mod_gradc
            d_nz[d_idx] = d_cz[d_idx] * mod_gradc

            # discretized Dirac Delta function
            d_ddelta[d_idx] = 1./mod_gradc


class SY11DiracDelta(Equation):
    r"""Discretized dirac-delta for the SY11 formulation Eq. (14) in [SY11]

    This is essentially the same as computing the color gradient, the
    only difference being that this might be called with a reduced
    smoothing length.

    Note that the normals should be computed using the
    SY11ColorGradient equation. This function will effectively
    overwrite the color gradient.

    """

    def __init__(self, dest, sources, epsilon=1e-6):
        self.epsilon2 = epsilon*epsilon
        super(SY11DiracDelta, self).__init__(dest, sources)

    def initialize(self, d_idx, d_cx, d_cy, d_cz, d_ddelta):
        # color gradient
        d_cx[d_idx] = 0.0
        d_cy[d_idx] = 0.0
        d_cz[d_idx] = 0.0

        # discretized dirac delta
        d_ddelta[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_color, s_color, d_cx, d_cy, d_cz,
             d_V, s_V, DWIJ):

        # average particle volume
        psiab1 = 2.0/(d_V[d_idx] + s_V[s_idx])

        # difference in color divided by psiab. Eq. (13) in [SY11]
        Cba = (s_color[s_idx] - d_color[d_idx]) * psiab1

        # color gradient
        d_cx[d_idx] += Cba * DWIJ[0]
        d_cy[d_idx] += Cba * DWIJ[1]
        d_cz[d_idx] += Cba * DWIJ[2]

    def post_loop(self, d_idx, d_cx, d_cy, d_cz,
                  d_nx, d_ny, d_nz, d_N, d_ddelta):
        # absolute value of the color gradient
        mod_gradc2 = d_cx[d_idx]*d_cx[d_idx] + \
            d_cy[d_idx]*d_cy[d_idx] + \
            d_cz[d_idx]*d_cz[d_idx]

        # avoid sqrt computations on non-interface particles
        if mod_gradc2 > self.epsilon2:
            mod_gradc = sqrt(mod_gradc2)

            # discretized Dirac Delta function
            d_ddelta[d_idx] = mod_gradc


class InterfaceCurvatureFromNumberDensity(Equation):
    r"""Interface curvature using number density. Eq. (15) in [SY11]:

    .. math::

        \kappa_a = \sum_b \frac{2.0}{\psi_a + \psi_b}
        \left(\boldsymbol{n_a} - \boldsymbol{n_b}\right) \cdot
        \nabla_a W_{ab}

    """

    def __init__(self, dest, sources, with_morris_correction=True):
        self.with_morris_correction = with_morris_correction

        super(InterfaceCurvatureFromNumberDensity, self).__init__(dest,
                                                                  sources)

    def initialize(self, d_idx, d_kappa, d_wij_sum):
        d_kappa[d_idx] = 0.0
        d_wij_sum[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_kappa, d_nx, d_ny, d_nz, s_nx, s_ny, s_nz,
             d_V, s_V, d_N, s_N, d_wij_sum, s_rho, s_m, WIJ, DWIJ):

        nijdotdwij = (d_nx[d_idx] - s_nx[s_idx]) * DWIJ[0] + \
            (d_ny[d_idx] - s_ny[s_idx]) * DWIJ[1] + \
            (d_nz[d_idx] - s_nz[s_idx]) * DWIJ[2]

        # averaged particle number density
        psiij1 = 2.0/(d_V[d_idx] + s_V[s_idx])

        # local number density with reliable normals Eq. (24) in [JM00]
        tmp = 1.0
        if self.with_morris_correction:
            tmp = min(d_N[d_idx], s_N[s_idx])

        d_wij_sum[d_idx] += tmp * s_m[s_idx]/s_rho[s_idx] * WIJ

        # Eq. (15) in [SY11] with correction Eq. (22) in [JM00]
        d_kappa[d_idx] += tmp * psiij1 * nijdotdwij

    def post_loop(self, d_idx, d_wij_sum, d_nx, d_kappa):
        # correct the curvature estimate. Eq. (23) in [JM00]
        if self.with_morris_correction:
            if d_wij_sum[d_idx] > 1e-12:
                d_kappa[d_idx] /= d_wij_sum[d_idx]


class ShadlooYildizSurfaceTensionForce(Equation):
    r"""Acceleration due to surface tension force Eq. (7,9) in [SY11]:

    .. math:

        \frac{d\boldsymbol{v}_a} = \frac{1}{m_a} \sigma \kappa_a
        \boldsymbol{n}_a \delta_a^s\,,

    where, :math:`\delta^s` is the discretized dirac delta function,
    :math:`\boldsymbol{n}` is the interface normal, :math:`\kappa` is
    the discretized interface curvature and :math:`\sigma` is the
    surface tension force constant.

    """

    def __init__(self, dest, sources, sigma=0.1):
        self.sigma = sigma

        # base class initialization
        super(ShadlooYildizSurfaceTensionForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_au, d_av, d_aw, d_kappa,
             d_nx, d_ny, d_nz, d_m, d_rho, d_ddelta):

        mi = 1./d_m[d_idx]
        rhoi = 1./d_rho[d_idx]

        # acceleration per uint mass term Eq. (7) in [SY11]
        tmp = self.sigma * d_kappa[d_idx] * d_ddelta[d_idx] * rhoi

        d_au[d_idx] += tmp * d_nx[d_idx]
        d_av[d_idx] += tmp * d_ny[d_idx]
        d_aw[d_idx] += tmp * d_nz[d_idx]


class CSFSurfaceTensionForce(Equation):
    r"""Acceleration due to surface tension force Eq. (25) in [JM00]:

    .. math:

        \frac{d\boldsymbol{v}_a}{dt} = \frac{1}{rho_a} \sigma_b \kappa_a
        \boldsymbol{n}_a

    Note that as per Eq. (17) in [JM00], the un-normalized normal is
    basically the gradient of the color function. The acceleration
    term therefore depends on the gradient of the color field.

    """

    def __init__(self, dest, sources, sigma=0.1):
        self.sigma = sigma

        # base class initialization
        super(CSFSurfaceTensionForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_au, d_av, d_aw, d_kappa,
             d_cx, d_cy, d_cz, d_rho):

        rhoi = 1./d_rho[d_idx]

        # acceleration per uint mass term Eq. (25) in [JM00]
        tmp = self.sigma * d_kappa[d_idx] * rhoi

        d_au[d_idx] += tmp * d_cx[d_idx]
        d_av[d_idx] += tmp * d_cy[d_idx]
        d_aw[d_idx] += tmp * d_cz[d_idx]


class AdamiReproducingDivergence(Equation):
    r"""Reproducing divergence approximation Eq. (20) in [A10] to
    compute the curvature

    .. math::

        \nabla \cdot \boldsymbol{\phi}_a = d\frac{\sum_b
        \boldsymbol{\phi}_{ab}\cdot \nabla_a
        W_{ab}V_b}{\sum_b\boldsymbol{x}_{ab}\cdot \nabla_a W_{ab} V_b}

    """

    def __init__(self, dest, sources, dim):
        self.dim = dim
        super(AdamiReproducingDivergence, self).__init__(dest, sources)

    def initialize(self, d_idx, d_kappa, d_wij_sum):
        d_kappa[d_idx] = 0.0
        d_wij_sum[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_kappa, d_wij_sum,
             d_nx, d_ny, d_nz, s_nx, s_ny, s_nz, d_V, s_V,
             DWIJ, XIJ, RIJ, EPS, d_N, s_N, d_color, s_color):
        # particle volumes
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]

        color_diff = abs(d_color[d_idx] - s_color[s_idx])

        if color_diff == 1.0:
            phi_ij = -1.0
        else:
            phi_ij = 1.0
        # dot product in the numerator of Eq. (20)
        nijdotdwij = (d_nx[d_idx] - phi_ij * s_nx[s_idx]) * DWIJ[0] + \
            (d_ny[d_idx] - phi_ij * s_ny[s_idx]) * DWIJ[1] + \
            (d_nz[d_idx] - phi_ij * s_nz[s_idx]) * DWIJ[2]

        # dot product in the denominator of Eq. (20)
        xijdotdwij = XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] + XIJ[2]*DWIJ[2]

        tmp = min(d_N[d_idx], s_N[s_idx])

        # accumulate the contributions
        d_kappa[d_idx] += tmp * nijdotdwij * Vj
        d_wij_sum[d_idx] += tmp * xijdotdwij * Vj

    def post_loop(self, d_idx, d_kappa, d_wij_sum):
        # normalize the curvature estimate
        if abs(d_wij_sum[d_idx]) > 1e-12:
            d_kappa[d_idx] /= d_wij_sum[d_idx]
        d_kappa[d_idx] *= self.dim


class CSFSurfaceTensionForceAdami(Equation):
    def __init__(self, dest, sources, sigma):
        self.sigma = sigma
        super(CSFSurfaceTensionForceAdami, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def post_loop(self, d_idx, d_au, d_av, d_aw, d_kappa, d_cx, d_cy, d_cz,
                  d_m, d_alpha, d_rho):
        d_au[d_idx] += -self.sigma * d_kappa[d_idx] * d_cx[d_idx]/d_rho[d_idx]
        d_av[d_idx] += -self.sigma * d_kappa[d_idx] * d_cy[d_idx]/d_rho[d_idx]
        d_aw[d_idx] += -self.sigma * d_kappa[d_idx] * d_cz[d_idx]/d_rho[d_idx]


class ShadlooViscosity(Equation):
    def __init__(self, dest, sources, alpha):
        self.alpha = alpha
        super(ShadlooViscosity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_au, d_av, d_aw, d_h, s_idx, s_h, d_cs, s_cs, d_rho,
             s_rho, VIJ, XIJ, d_V, s_V, R2IJ, EPS, DWIJ):
        mu1 = 0.125*self.alpha*d_h[d_idx]*d_cs[d_idx]*d_rho[d_idx]
        mu2 = 0.125*self.alpha*s_h[s_idx]*s_cs[s_idx]*s_rho[s_idx]
        mu12 = 2.0*mu1*mu2/(mu1 + mu2)
        vijdotxij = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]
        denominator = d_V[d_idx]*s_V[s_idx]*(R2IJ + EPS)
        piij = 8.0*mu12*vijdotxij/denominator
        d_au[d_idx] += -piij*DWIJ[0]
        d_av[d_idx] += -piij*DWIJ[1]
        d_aw[d_idx] += -piij*DWIJ[2]


class AdamiColorGradient(Equation):

    r"""Gradient of color Eq. (14) in [A10]

    .. math::

        \nabla c_a = \frac{1}{V_a}\sum_b \left[V_a^2 + V_b^2
        \right]\tilde{c}_{ab}\nabla_a W_{ab}\,,

    where, the average :math:`\tilde{c}_{ab}` is defined as

    .. math::

        \tilde{c}_{ab} = \frac{\rho_b}{\rho_a + \rho_b}c_a +
        \frac{\rho_a}{\rho_a + \rho_b}c_b

    """

    def initialize(self, d_idx, d_cx, d_cy, d_cz, d_nx, d_ny, d_nz, d_ddelta,
                   d_N):
        d_cx[d_idx] = 0.0
        d_cy[d_idx] = 0.0
        d_cz[d_idx] = 0.0

        d_nx[d_idx] = 0.0
        d_ny[d_idx] = 0.0
        d_nz[d_idx] = 0.0

        # reliability indicator for normals
        d_N[d_idx] = 0.0

        # Discretized dirac-delta
        d_ddelta[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_V, s_V, d_rho, s_rho,
             d_cx, d_cy, d_cz, d_color, s_color, DWIJ):

        # particle volumes
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # averaged particle color
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        rhoij1 = 1./(rhoi + rhoj)

        color_diff = abs((d_color[d_idx] - s_color[s_idx]))
        # Eq. (15) in [A10]

        if color_diff == 0.0:
            cij = 0.0
        else:
            cij = rhoj * rhoij1 * 0.0 + rhoi * rhoij1 * 1.0


        # comute the gradient
        tmp = cij * (Vi2 + Vj2)/Vi

        d_cx[d_idx] += tmp * DWIJ[0]
        d_cy[d_idx] += tmp * DWIJ[1]
        d_cz[d_idx] += tmp * DWIJ[2]

    def post_loop(self, d_idx, d_cx, d_cy, d_cz, d_h,
                  d_nx, d_ny, d_nz, d_ddelta, d_N):
        # absolute value of the color gradient
        mod_gradc2 = d_cx[d_idx]*d_cx[d_idx] + \
            d_cy[d_idx]*d_cy[d_idx] + \
            d_cz[d_idx]*d_cz[d_idx]

        # avoid sqrt computations on non-interface particles
        h2 = d_h[d_idx]*d_h[d_idx]
        if mod_gradc2 > 0.0:
            # this normal is reliable in the sense of [JM00]
            d_N[d_idx] = 1.0

            # compute the normals
            one_mod_gradc = 1./sqrt(mod_gradc2)

            d_nx[d_idx] = d_cx[d_idx] * one_mod_gradc
            d_ny[d_idx] = d_cy[d_idx] * one_mod_gradc
            d_nz[d_idx] = d_cz[d_idx] * one_mod_gradc

            # discretized dirac delta
            d_ddelta[d_idx] = 1./one_mod_gradc


def get_surface_tension_equations(fluids, solids, scheme, rho0, p0, c0, b,
                                  factor1, factor2, nu, sigma, d, epsilon,
                                  gamma, real=False):
    """This function returns the required equations for the multiphase formulation
    taking inputs of the fluid particles array, solid particles array, the
    scheme to be used and other physical parameters

    Parameters
    -----------

    fluids: list
        List of names of fluid particle arrays
    solids: list
        List of names of solid particle arrays
    scheme: string
        The scheme with which the equations are to be setup.
        Supported Schemes:

        1. TVF scheme with Morris' surface tension.
           String to be used: "tvf"
        2. Adami's surface tension implementation which doesn't involve
           calculation of curvature. String to be used: "adami_stress"
        3. Adami's surface tension implementation which involves
           calculation of curvature. String to be used: "adami"
        4. Shadloo Yildiz surface tension formulation.
           String to be used: "shadloo"
        5. Morris' surface tension formulation. This is the default scheme
           which will be used if none of the above strings are input as
           scheme.

    rho0 : float
        The reference density of the medium (Currently multiple reference
        densities for different particles is not supported)
    p0 : float
        The background pressure of the medium(Currently multiple background
        pressures for different particles is not supported)
    c0 : float
        The speed of sound of the medium(Currently multiple speeds of sounds
        for different particles is not supported)
    b : float
        The b parameter of the generalized Tait Equation of State. Refer to
        the Tait Equation's documentation for reference
    factor1 : float
        The factor for scaling of smoothing length for calculation of
        interface curvature number for shadloo's scheme
    factor2 : float
        The factor for scaling back of smoothing length for calculation of
        forces after calculating the interface curvature number in shadloo's
        scheme
    nu : float
        The kinematic viscosity of the medium
    sigma : float
        The surface tension of the system
    d : int
        The number of dimensions of the problem in the cartesian space
    epsilon: float
        Put this option false if the equations are supposed to be evaluated
        for the ghost particles, else keep it True

    """
    if scheme == 'tvf':
        result = []
        equations = []
        for i in fluids+solids:
            equations.append(SummationDensity(dest=i, sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(StateEquation(dest=i, sources=None, rho0=rho0,
                                           p0=p0))
            equations.append(SmoothedColor(dest=i, sources=fluids+solids))
        for i in solids:
            equations.append(SolidWallPressureBCnoDensity(dest=i,
                                                          sources=fluids))
            equations.append(SmoothedColor(dest=i, sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(MorrisColorGradient(dest=i, sources=fluids+solids,
                                                 epsilon=epsilon))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(InterfaceCurvatureFromNumberDensity(
                dest=i, sources=fluids+solids, with_morris_correction=True))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(MomentumEquationPressureGradient(dest=i,
                             sources=fluids+solids, pb=p0))
            equations.append(MomentumEquationViscosity(dest=i, sources=fluids,
                             nu=nu))
            equations.append(CSFSurfaceTensionForce(dest=i, sources=None,
                             sigma=sigma))
            equations.append(MomentumEquationArtificialStress(dest=i,
                                                              sources=fluids))
            if len(solids) != 0:
                equations.append(SolidWallNoSlipBC(dest=i, sources=solids,
                                 nu=nu))
        result.append(Group(equations))
    elif scheme == 'adami_stress':
        result = []
        equations = []
        for i in fluids+solids:
            equations.append(SummationDensitySourceMass(dest=i,
                                                        sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(TaitEOS(dest=i, sources=None, c0=c0,
                             gamma=gamma, p0=p0, rho0=rho0))
        for i in solids:
            equations.append(SolidWallPressureBCnoDensity(dest=i,
                             sources=fluids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(ColorGradientAdami(dest=i, sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(ConstructStressMatrix(dest=i, sources=None,
                                                   sigma=sigma, d=d))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(MomentumEquationPressureGradientHuAdams(dest=i,
                             sources=fluids+solids))
            equations.append(MomentumEquationViscosityAdami(dest=i,
                                                            sources=fluids))
            equations.append(SurfaceForceAdami(dest=i, sources=fluids+solids))
            if len(solids) != 0:
                equations.append(SolidWallNoSlipBC(dest=i, sources=solids,
                                                   nu=nu))
        result.append(Group(equations))
    elif scheme == 'adami':
        result = []
        equations = []
        for i in fluids+solids:
            equations.append(SummationDensitySourceMass(dest=i,
                                                        sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(TaitEOS(dest=i, sources=None, c0=c0,
                             gamma=gamma, p0=p0, rho0=rho0))
        for i in solids:
            equations.append(SolidWallPressureBCnoDensity(dest=i,
                                                          sources=fluids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(AdamiColorGradient(dest=i, sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(AdamiReproducingDivergence(dest=i,
                                                        sources=fluids+solids,
                                                        dim=d))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(MomentumEquationPressureGradientAdami(
                dest=i, sources=fluids+solids))
            equations.append(MomentumEquationViscosityAdami(dest=i,
                                                            sources=fluids))
            equations.append(CSFSurfaceTensionForceAdami(dest=i,
                                                         sources=None,
                                                         sigma=sigma))
            if len(solids) != 0:
                equations.append(SolidWallNoSlipBC(dest=i, sources=solids,
                                 nu=nu))
        result.append(Group(equations))
    elif scheme == 'shadloo':
        result = []
        equations = []
        for i in fluids+solids:
            equations.append(SummationDensitySourceMass(dest=i,
                                                        sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(IsothermalEOS(dest=i, sources=None,
                                           p0=p0, c0=c0, rho0=rho0))
            equations.append(SY11ColorGradient(dest=i, sources=fluids+solids))
        for i in solids:
            equations.append(SolidWallPressureBCnoDensity(dest=i,
                                                          sources=fluids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(ScaleSmoothingLength(dest=i, sources=None,
                                                  factor=factor1))
        result.append(Group(equations, real=real, update_nnps=True))
        equations = []
        for i in fluids:
            equations.append(SY11DiracDelta(dest=i, sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(InterfaceCurvatureFromNumberDensity(
                dest=i, sources=fluids+solids, with_morris_correction=True))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(ScaleSmoothingLength(dest=i, sources=None,
                                                  factor=factor2))
        result.append(Group(equations, real=real, update_nnps=True))
        equations = []
        for i in fluids:
            equations.append(MomentumEquationPressureGradient(
                dest=i, sources=fluids+solids, pb=0.0))
            equations.append(MomentumEquationViscosity(dest=i, sources=fluids,
                                                       nu=nu))
            equations.append(ShadlooYildizSurfaceTensionForce(dest=i,
                                                              sources=None,
                                                              sigma=sigma))
            if len(solids) != 0:
                equations.append(SolidWallNoSlipBC(dest=i, sources=solids,
                                                   nu=nu))
        result.append(Group(equations))
    else:
        result = []
        equations = []
        for i in fluids+solids:
            equations.append(SummationDensitySourceMass(dest=i,
                                                        sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(TaitEOS(dest=i, sources=None, rho0=rho0, c0=c0,
                                     gamma=gamma, p0=p0))
            equations.append(SmoothedColor(dest=i, sources=fluids+solids))
        for i in solids:
            equations.append(SolidWallPressureBCnoDensity(dest=i,
                                                          sources=fluids))
            equations.append(SmoothedColor(dest=i, sources=fluids+solids))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(MorrisColorGradient(dest=i, sources=fluids+solids,
                                                 epsilon=epsilon))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(InterfaceCurvatureFromDensity(
                dest=i, sources=fluids+solids, with_morris_correction=True))
        result.append(Group(equations, real=real))
        equations = []
        for i in fluids:
            equations.append(MomentumEquationPressureGradientMorris(dest=i,
                             sources=fluids+solids))
            equations.append(MomentumEquationViscosityMorris(dest=i,
                                                             sources=fluids))
            equations.append(CSFSurfaceTensionForce(dest=i, sources=None,
                                                    sigma=sigma))
            if len(solids) != 0:
                equations.append(SolidWallNoSlipBC(dest=i, sources=solids,
                                                   nu=nu))
        result.append(Group(equations))
    return result



import pytest
import numpy as np

from pysph.base.utils import get_particle_array  # noqa: E402
from pysph.base.device_helper import DeviceHelper
from compyle.config import get_config
import compyle.array as array


def setup_module():
    get_config().use_openmp = True


def teardown_module():
    get_config().use_openmp = False


def check_import(backend):
    if backend == 'opencl':
        pytest.importorskip('pyopencl')
    if backend == 'cuda':
        pytest.importorskip('pycuda')


check_all_backends = pytest.mark.parametrize('backend',
                                            ['cython', 'opencl', 'cuda'])


class TestDeviceHelper(object):
    def setup(self):
        self.pa = get_particle_array(name='f', x=[0.0, 1.0], m=1.0, rho=2.0)

    @check_all_backends
    def test_simple(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)

        # Then
        assert np.allclose(pa.x, h.x.get())
        assert np.allclose(pa.y, h.y.get())
        assert np.allclose(pa.m, h.m.get())
        assert np.allclose(pa.rho, h.rho.get())
        assert np.allclose(pa.tag, h.tag.get())

    @check_all_backends
    def test_push_correctly_sets_values_with_args(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)
        assert pa.tag[0] == 0

        # When
        pa.set_device_helper(h)
        pa.x[:] = [2.0, 3.0]
        pa.rho[0] = 1.0
        pa.tag[:] = 1
        h.push('x', 'rho', 'tag')

        # Then
        assert np.allclose(pa.x, h.x.get())
        assert np.allclose(pa.y, h.y.get())
        assert np.allclose(pa.m, h.m.get())
        assert np.allclose(pa.rho, h.rho.get())
        assert np.allclose(pa.tag, h.tag.get())

    @check_all_backends
    def test_push_correctly_sets_values_with_no_args(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        pa.x[:] = 1.0
        pa.rho[:] = 1.0
        pa.m[:] = 1.0
        pa.tag[:] = [1, 2]
        h.push()

        # Then
        assert np.allclose(pa.x, h.x.get())
        assert np.allclose(pa.y, h.y.get())
        assert np.allclose(pa.m, h.m.get())
        assert np.allclose(pa.rho, h.rho.get())
        assert np.allclose(pa.tag, h.tag.get())

    @check_all_backends
    def test_pull_correctly_sets_values_with_args(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)
        assert pa.tag[0] == 0

        # When
        pa.set_device_helper(h)
        h.x.set(np.array([2.0, 3.0], h.x.dtype))
        h.rho[0] = 1.0
        h.tag[:] = 1
        h.pull('x', 'rho', 'tag')

        # Then
        assert np.allclose(pa.x, h.x.get())
        assert np.allclose(pa.y, h.y.get())
        assert np.allclose(pa.m, h.m.get())
        assert np.allclose(pa.rho, h.rho.get())
        assert np.allclose(pa.tag, h.tag.get())

    @check_all_backends
    def test_pull_correctly_sets_values_with_no_args(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        h.x[:] = 1.0
        h.rho[:] = 1.0
        h.m[:] = 1.0
        h.tag[:] = np.array([1, 2], h.tag.dtype)
        h.pull()

        # Then
        assert np.allclose(pa.x, h.x.get())
        assert np.allclose(pa.y, h.y.get())
        assert np.allclose(pa.m, h.m.get())
        assert np.allclose(pa.rho, h.rho.get())
        assert np.allclose(pa.tag, h.tag.get())

    @check_all_backends
    def test_max_provides_maximum(self, backend):
        check_import(backend)
        self.setup()
        # Given/When
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # Then
        assert h.max('x') == 1.0

    @check_all_backends
    def test_that_adding_removing_prop_to_array_updates_gpu(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        pa.add_property('test', data=[3.0, 4.0])

        # Then
        assert np.allclose(pa.test, h.test.get())

        # When
        pa.remove_property('test')

        # Then
        assert not hasattr(h, 'test')
        assert 'test' not in h._data
        assert 'test' not in h.properties

    @check_all_backends
    def test_resize_works(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        pa.extend(2)
        pa.align_particles()
        h.resize(4)

        # Then
        assert np.allclose(pa.x[:2], h.x[:2].get())
        assert np.allclose(pa.m[:2], h.m[:2].get())
        assert np.allclose(pa.rho[:2], h.rho[:2].get())
        assert np.allclose(pa.tag[:2], h.tag[:2].get())

        # When
        pa.remove_particles([2, 3])
        pa.align_particles()
        old_x = h.x.data
        h.resize(2)

        # Then
        assert old_x == h.x.data
        assert np.allclose(pa.x, h.x.get())
        assert np.allclose(pa.y, h.y.get())
        assert np.allclose(pa.m, h.m.get())
        assert np.allclose(pa.rho, h.rho.get())
        assert np.allclose(pa.tag, h.tag.get())

    @check_all_backends
    def test_get_number_of_particles(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        h.resize(5)
        h.x.set(np.array([2.0, 3.0, 4.0, 5.0, 6.0], h.x.dtype))
        h.tag.set(np.array([0, 0, 1, 0, 1], h.tag.dtype))

        h.align_particles()

        # Then
        assert h.get_number_of_particles() == 5
        assert h.get_number_of_particles(real=True) == 3

    @check_all_backends
    def test_align(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        pa.add_property('force', stride=3)
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        n = 5
        h.resize(n)
        h.x.set(np.array([2.0, 3.0, 4.0, 5.0, 6.0], h.x.dtype))
        h.force.set(np.arange(n*3, dtype=h.force.dtype))

        indices = array.arange(4, -1, -1, dtype=np.int32,
                               backend=backend)

        h.align(indices)

        # Then
        assert np.all(h.x.get() == np.array([6., 5., 4., 3., 2.]))
        x = np.arange(n*3)
        x.shape = (n, 3)
        expect = x[::-1, :].ravel()
        assert np.all(h.force.get() == expect)

    @check_all_backends
    def test_align_particles(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        h.resize(5)
        h.x.set(np.array([2.0, 3.0, 4.0, 5.0, 6.0], h.x.dtype))
        h.tag.set(np.array([0, 0, 1, 0, 1], h.tag.dtype))

        h.align_particles()

        # Then
        x = h.x.get()
        assert np.all(np.sort(x[:-2]) == np.array([2., 3., 5.]))

    @check_all_backends
    def test_remove_particles(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        h.resize(4)
        h.x.set(np.array([2.0, 3.0, 4.0, 5.0], h.x.dtype))

        indices = np.array([1, 2], dtype=np.uint32)
        indices = array.to_device(indices, backend=backend)

        h.remove_particles(indices)

        # Then
        assert np.all(np.sort(h.x.get()) == np.array([2., 5.]))

    @check_all_backends
    def test_remove_tagged_particles(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        h.resize(5)
        h.x.set(np.array([2.0, 3.0, 4.0, 5.0, 6.0], h.x.dtype))
        h.tag.set(np.array([0, 0, 1, 0, 1], h.tag.dtype))

        h.remove_tagged_particles(1)

        # Then
        assert np.all(np.sort(h.x.get()) == np.array([2., 3., 5.]))

    @check_all_backends
    def test_add_particles(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)
        x = array.zeros(4, np.float32, backend=backend)

        h.add_particles(x=x)

        # Then
        assert np.all(np.sort(h.x.get()) == np.array([0., 0., 0., 0., 0., 1.]))

    @check_all_backends
    def test_extend(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = self.pa
        h = DeviceHelper(pa, backend=backend)

        # When
        pa.set_device_helper(h)

        h.extend(4)

        # Then
        assert h.get_number_of_particles() == 6

    @check_all_backends
    def test_append_parray(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa1 = self.pa
        pa2 = get_particle_array(name='s', x=[0.0, 1.0], m=1.0, rho=2.0)
        h = DeviceHelper(pa1, backend=backend)
        pa1.set_device_helper(h)

        # When
        h.append_parray(pa2)

        # Then
        assert h.get_number_of_particles() == 4

    @check_all_backends
    def test_empty_clone(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = get_particle_array(name='f', x=[0.0, 1.0, 2.0, 3.0],
                                m=1.0, rho=2.0)
        h = DeviceHelper(pa, backend=backend)
        pa.set_device_helper(h)

        # When
        result_pa = h.empty_clone()

        # Then
        assert result_pa.gpu.get_number_of_particles() == 0
        assert result_pa.name == 'f'

    @check_all_backends
    def test_extract_particles(self, backend):
        check_import(backend)
        self.setup()
        # Given
        pa = get_particle_array(name='f', x=[0.0, 1.0, 2.0, 3.0],
                                m=1.0, rho=2.0)
        h = DeviceHelper(pa, backend=backend)
        pa.set_device_helper(h)

        # When
        indices = np.array([1, 2], dtype=np.uint32)
        indices = array.to_device(indices, backend=backend)

        result_pa = h.empty_clone()
        h.extract_particles(indices, result_pa)

        # Then
        assert result_pa.gpu.get_number_of_particles() == 2

    def test_update_minmax_cl(self):
        backend = 'opencl'
        check_import(backend)
        self.setup()

        # Given
        x = [0.0, -1.0, 2.0, 3.0]
        y = [0.0, 1.0, -2.0, 3.0]
        z = [0.0, 1.0, 2.0, -3.0]
        h = [4.0, 1.0, 2.0, 3.0]

        pa = get_particle_array(x=x, y=y, z=z, h=h)
        h = DeviceHelper(pa, backend=backend)
        pa.set_device_helper(h)

        # When
        h.update_minmax_cl(['x', 'y', 'z', 'h'])

        # Then
        assert h.x.minimum == -1.0
        assert h.x.maximum == 3.0

        assert h.y.minimum == -2.0
        assert h.y.maximum == 3.0

        assert h.z.minimum == -3.0
        assert h.z.maximum == 2.0

        assert h.h.minimum == 1.0
        assert h.h.maximum == 4.0

        # When
        h.x.maximum, h.x.minimum = 100., 100.
        h.y.maximum, h.y.minimum = 100., 100.
        h.z.maximum, h.z.minimum = 100., 100.
        h.h.maximum, h.h.minimum = 100., 100.

        h.update_minmax_cl(['x', 'y', 'z', 'h'], only_min=True)

        # Then
        assert h.x.minimum == -1.0
        assert h.x.maximum == 100.0

        assert h.y.minimum == -2.0
        assert h.y.maximum == 100.0

        assert h.z.minimum == -3.0
        assert h.z.maximum == 100.0

        assert h.h.minimum == 1.0
        assert h.h.maximum == 100.0

        # When
        h.x.maximum, h.x.minimum = 100., 100.
        h.y.maximum, h.y.minimum = 100., 100.
        h.z.maximum, h.z.minimum = 100., 100.
        h.h.maximum, h.h.minimum = 100., 100.

        h.update_minmax_cl(['x', 'y', 'z', 'h'], only_max=True)

        # Then
        assert h.x.minimum == 100.0
        assert h.x.maximum == 3.0

        assert h.y.minimum == 100.0
        assert h.y.maximum == 3.0

        assert h.z.minimum == 100.0
        assert h.z.maximum == 2.0

        assert h.h.minimum == 100.0
        assert h.h.maximum == 4.0



"""Cylindrical dam break over a sloping parabolic dry bed. (24 mins)

The case is described in "A corrected smooth particle hydrodynamics formulation
of the shallow-water equations", Miguel Rodriguez-Paz and Javier Bonet,
Computers & Structures, Vol 83, pp 1396-1410 (2005).
DOI:10.1016/j.compstruc.2004.11.025

"""
# Numpy
from numpy import (ones_like, zeros, pi, arange, concatenate, sin, cos)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        InitialGuessDensity, SummationDensity, DensityResidual,
        DensityNewtonRaphsonIteration, CheckConvergence, UpdateSmoothingLength,
        SWEOS, SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator


# Constants
rho_w = 1000.0
g = 9.81
dim = 2


class ParticleAccelerationAnalytBottom(ParticleAcceleration):
    def post_loop(self, d_y, d_idx, d_u, d_v, d_tu, d_tv, d_au, d_av, d_Sfx,
                  d_Sfy):
        # Gradient of fluid bottom
        bx = -0.839
        by = 0.909 * d_y[d_idx]

        # Curvature of fluid bottom
        bxx = 0
        bxy = 0
        byy = 0.909

        vikivi = (d_u[d_idx]*d_u[d_idx]*bxx
                  + 2*d_u[d_idx]*d_v[d_idx]*bxy
                  + d_v[d_idx]*d_v[d_idx]*byy)
        tidotgradbi = d_tu[d_idx]*bx + d_tv[d_idx]*by
        gradbidotgradbi = bx**2 + by**2
        temp3 = self.g + vikivi - tidotgradbi
        temp4 = 1 + gradbidotgradbi
        if not self.v_only:
            d_au[d_idx] = -(temp3/temp4)*bx - d_tu[d_idx] - d_Sfx[d_idx]
        if not self.u_only:
            d_av[d_idx] = -(temp3/temp4)*by - d_tv[d_idx] - d_Sfy[d_idx]


class CylindricalDamBreakSlopingBed(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=2.0,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dw0", action="store", type=float, dest="dw0", default=0.25,
            help="Initial depth of the fluid column (m)")
        group.add_argument(
            "--r", action="store", type=float, dest="r", default=0.25,
            help="Initial radius of the fluid column (m)")
        group.add_argument(
            "--n", action="store", type=float, dest="n", default=50,
            help="Number of concentric fluid particle circles (Determines\
            spacing btw particles, dr = r/n)")
        group.add_argument(
            "--R", action="store", type=float, dest="R", default=1.1,
            help="Radius of curvature of bed surface (m)")
        group.add_argument(
            "--theta", action="store", type=float, dest="theta", default=40,
            help="Bed slope measured along clockwise direction from horizontal\
            (degrees)")
        group.add_argument(
            "--xcen", action="store", type=float, dest="xcen", default=0.25,
            help="x-coordinate of the center of the base of the fluid column\
            (m)")
        group.add_argument(
            "--ycen", action="store", type=float, dest="ycen", default=0.5,
            help="y-coordinate of the center of the base of the fluid column\
            (m)")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dw0 = self.options.dw0
        self.r = self.options.r
        self.n = self.options.n
        self.R = self.options.R
        self.theta = self.options.theta
        self.xcen = self.options.xcen
        self.ycen = self.options.ycen

    def create_particles(self):
        n = self.n
        r = self.r
        dr = r / n

        d = self.dw0
        hdx = self.hdx

        x = zeros(0)
        y = zeros(0)

        # Create circular patch of fluid in a radial grid
        rad = 0.0
        for j in range(1, n+1):
                npnts = 4 * j
                dtheta = (2*pi) / npnts

                theta = arange(0, 2*pi-1e-10, dtheta)
                rad = rad + dr

                _x = rad * cos(theta)
                _y = rad * sin(theta)

                x = concatenate((x, _x))
                y = concatenate((y, _y))

        arr_ones = ones_like(x)

        m = arr_ones * (1.56*dr*dr) * rho_w * d

        rho = arr_ones * rho_w * d
        rho0 = arr_ones * rho_w * d

        h = arr_ones * hdx * dr
        h0 = arr_ones * hdx * dr

        # Analytially set fluid bottom height and gradients
        R = self.R
        theta = self.theta
        b = y**2/R + (theta*(pi/180.))*x
        bx = arr_ones * -0.839
        by = 0.909 * y
        byy = 0.909

        # Eccentricity in fluid column from mid section
        x += self.xcen
        y += self.ycen

        fluid = gpa_swe(x=x, y=y, m=m, rho=rho, rho0=rho0, h=h, h0=h0,
                        b=b, bx=bx, by=by, byy=byy, name='fluid')

        compute_initial_props([fluid])
        return [fluid]

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = 1.0
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.2,
            adaptive_timestep=True,
            output_at_times=(0.2, 0.5, 1.0),
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    InitialGuessDensity(dim=dim, dest='fluid',
                                        sources=['fluid']),
                    UpdateSmoothingLength(dim=dim, dest='fluid')
                ], update_nnps=True
            ),

            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                            dest='fluid', sources=['fluid']),
                    SummationDensity(dest='fluid', sources=['fluid']),
                    DensityResidual('fluid')
                ]
            ),

            Group(
                equations=[
                    Group(
                        equations=[
                            DensityNewtonRaphsonIteration(dim=dim, 
                                                          dest='fluid'),
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                        ], update_nnps=True
                    ),

                    Group(
                        equations=[
                            CorrectionFactorVariableSmoothingLength(
                                    dest='fluid', sources=['fluid']),
                            SummationDensity(dest='fluid', sources=['fluid']),
                            DensityResidual(dest='fluid'),
                            CheckConvergence(dest='fluid')
                        ],
                    )
                ], iterate=True, max_iterations=100
            ),

            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(dest='fluid',
                                                            sources=['fluid']),
                    SWEOS(dest='fluid')
                ]
            ),

            Group(
                equations=[
                    ParticleAccelerationAnalytBottom(dim=dim, dest='fluid', 
                                                     sources=['fluid'])
                ]
            )
        ]
        return equations

def compute_initial_props(particles):
    one_time_equations = [
        Group(
            equations=[
                CorrectionFactorVariableSmoothingLength(dest='fluid', 
                                                        sources=['fluid']),
                SWEOS(dest='fluid')
            ]
        ),
    ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = CylindricalDamBreakSlopingBed()
    app.run()



"""General post-processing utility for solution data"""

TVTK = True
try:
    from tvtk.api import tvtk, write_data
except (ImportError, SystemExit):
    TVTK = False

if TVTK:
    from tvtk.array_handler import array2vtk

from os import path
import numpy as np
import pysph.solver.utils as utils


def get_ke_history(files, array_name):
    t, ke = [], []
    for sd, array in utils.iter_output(files, array_name):
        t.append(sd['t'])
        m, u, v, w = array.get('m', 'u', 'v', 'w')
        _ke = 0.5 * np.sum( m * (u**2 + v**2 + w**2) )
        ke.append(_ke)
    return np.asarray(t), np.asarray(ke)


class Results(object):
    def __init__(self, dirname=None, fname=None, endswith=".npz"):
        self.dirname = dirname
        self.fname = fname
        self.endswith = endswith

        # the starting file number
        self.start = 0

        if ( (dirname is not None) and (fname is not None) ):
            self.load()

    def set_dirname(self, dirname):
        self.dirname=dirname

    def set_fname(self, fname):
        self.fname = fname

    def load(self):
        self.files = files = utils.get_files(
            self.dirname, self.fname, self.endswith)

        self.nfiles = len(files)

    def reload(self):
        self.start = self.nfiles
        self.load()

    def get_ke_history(self, array_name):
        self.t, self.ke = get_ke_history(self.files, array_name)

    def _write_vtk_snapshot(self, mesh, directory, _fname):
        fname = path.join(directory, _fname)
        write_data( mesh, fname )

    def write_vtk(self, array_name, props):
        if not TVTK:
            return

        # create a list of props
        if type(props) != list:
            props = [ props ]

        # create an output folder for the vtk files
        dirname = path.join(self.dirname, 'vtk')
        utils.mkdir(dirname)

        nfiles = self.nfiles
        for i in range(self.start, nfiles):
            f = self.files[i]
            data = utils.load(f)

            array = data['arrays'][array_name]
            num_particles = array.num_real_particles

            # save the points
            points = np.zeros( shape=(num_particles,3) )
            points[:, 0] = array.z
            points[:, 1] = array.y
            points[:, 2] = array.x

            mesh = tvtk.PolyData(points=points)

            # add the scalar props
            for prop in props:
                if prop == 'vmag':
                    u, v, w = array.get('u','v','w')
                    numpy_array = np.sqrt(u**2 + v**2 + w**2)
                else:
                    numpy_array = array.get(prop)

                vtkarray = array2vtk(numpy_array)
                vtkarray.SetName(prop)

                # add the array as point data
                mesh.point_data.add_array(vtkarray)

            # set the last prop as the active scalar
            mesh.point_data.set_active_scalars(props[-1])

            # spit it out
            fileno = data['solver_data']['count']
            _fname = self.fname + '_%s_%s'%(array_name, fileno)

            self._write_vtk_snapshot(mesh, dirname, _fname)

class PySPH2VTK(object):
    """Convert PySPH array data to Paraview legible VTK data"""
    def __init__(self, arrays, dirname='.', fileno=None):
        self.arrays = arrays
        self.dirname = dirname
        self.fileno=fileno

        array_dict = {}
        for array in arrays:
            array_dict[ array.name ] = array

        self.array_dict = array_dict

    def _write_vtk_snapshot(self, mesh, directory, _fname):
        fname = path.join(directory, _fname)
        write_data( mesh, fname )

    def write_vtk(self, array_name, props):
        # ceck if it is possible
        if not TVTK:
            raise RuntimeError('Cannot generate VTK output!')

        # check if the array is legal
        if array_name not in list(self.array_dict.keys()):
            raise RuntimeError('Array %s not defined'%array_name)

        # create a list of props
        if type(props) != list:
            props = [ props ]

        # create an output folder for the vtk files
        dirname = path.join(self.dirname, 'vtk')
        utils.mkdir(dirname)

        array = self.array_dict[array_name]
        num_particles = array.num_real_particles

        # save the points
        points = np.zeros( shape=(num_particles,3) )
        points[:, 0] = array.z
        points[:, 1] = array.y
        points[:, 2] = array.x

        mesh = tvtk.PolyData(points=points)

        # add the scalar props
        for prop in props:
            if prop == 'vmag':
                u, v, w = array.get('u','v','w')
                numpy_array = np.sqrt(u**2 + v**2 + w**2)
            else:
                numpy_array = array.get(prop)

            vtkarray = array2vtk(numpy_array)
            vtkarray.SetName(prop)

            # add the array as point data
            mesh.point_data.add_array(vtkarray)

        # set the last prop as the active scalar
        mesh.point_data.set_active_scalars(props[-1])

        # spit it out
        if self.fileno is None:
            _fname = '%s'%(array_name)
        else:
            _fname = '%s_%03d'%(array_name, self.fileno)

        self._write_vtk_snapshot(mesh, dirname, _fname)



import numpy as np
from pysph.tools.geometry import get_2d_wall


def get_beach_geometry_2d(dx=0.1, l=3.0, h=1.0, flat_l=1.0, angle=45.0,
                          num_layers=3):
    """
    Generates a beach like geometry which is commonly used for simulations
    related to SPHysics.

    Parameters
    ----------
    dx : Spacing between the particles
    l : Total length of the beach
    h : Height of the wall used at the beach position
    flat_l : Length of the flat part
    angle : Angle of the inclined part
    num_layers : number of layers

    Returns
    -------
    x : 1d numpy array with x coordinates of the beach
    y : 1d numpy array with y coordinates of the beach
    x4 : 1d numpy array with x coordinates of the obstacle
    y4 : 1d numpy array with y coordinates of the obstacle
    """

    theta = np.pi * angle / 180.0
    x1, y1 = get_2d_wall(dx, np.array([(flat_l + dx) / 2.0, 0.]), flat_l,
                         num_layers, False)
    x2 = np.arange(flat_l - l, 0.0, dx * np.cos(theta))
    h2 = (l - flat_l) * np.tan(theta)
    y2_layer = x2 * np.tan(-theta)
    x2 = np.tile(x2, num_layers)
    y2 = []
    for i in range(num_layers):
        y2.append(y2_layer - i * dx)
    y2 = np.ravel(np.array(y2))
    y3 = np.arange(h2 + dx, h + h2, dx)
    x3_layer = np.ones_like(y3) * (flat_l - l)
    y3 = np.tile(y3, num_layers)
    x3 = []
    for i in range(num_layers):
        x3.append(x3_layer - i * dx)
    x3 = np.ravel(np.array(x3))
    x = np.concatenate([x1, x2, x3])
    y = np.concatenate([y1, y2, y3])
    y4 = np.arange(dx, 2.0 * h, dx)
    x4_layer = np.ones_like(y4) * flat_l
    y4 = np.tile(y4, num_layers)
    x4 = []
    for i in range(num_layers):
        x4.append(x4_layer + i * dx)
    x4 = np.ravel(np.array(x4))
    return x, y, x4, y4



"""Poiseuille flow using the transport velocity formulation (5 minutes).
"""
import os

# numpy
import numpy as np

# PySPH imports
from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.solver.utils import load

from pysph.sph.scheme import TVFScheme


# Numerical setup
dx = 1.0/60.0
ghost_extent = 5 * dx
hdx = 1.0

# adaptive time steps
h0 = hdx * dx

class PoiseuilleFlow(Application):
    def initialize(self):
        self.d = 0.5
        self.Ly = 2*self.d
        self.Lx = 0.4*self.Ly
        self.rho0 = 1.0
        self.nu = 0.01

    def add_user_options(self, group):
        group.add_argument(
            "--re", action="store", type=float, dest="re", default=0.0125,
            help="Reynolds number of flow."
        )
        group.add_argument(
            "--remesh", action="store", type=float, dest="remesh", default=0,
            help="Remeshing frequency (setting it to zero disables it)."
        )

    def consume_user_options(self):
        self.re = self.options.re
        self.Vmax = self.nu*self.re/(2*self.d)
        self.c0 = 10*self.Vmax
        self.p0 = self.c0**2*self.rho0

        # The body force is adjusted to give the Required Reynold's number
        # based on the steady state maximum velocity Vmax:
        # Vmax = fx/(2*nu)*(d^2) at the centerline

        self.fx = self.Vmax * 2*self.nu/(self.d**2)
        # Setup default parameters.
        dt_cfl = 0.25 * h0/( self.c0 + self.Vmax )
        dt_viscous = 0.125 * h0**2/self.nu
        dt_force = 0.25 * np.sqrt(h0/self.fx)

        self.dt = min(dt_cfl, dt_viscous, dt_force)

    def configure_scheme(self):
        tf = 100.0
        scheme = self.scheme
        scheme.configure(c0=self.c0, p0=self.p0, pb=self.p0, gx=self.fx)
        scheme.configure_solver(tf=tf, dt=self.dt, pfreq=1000)
        print("dt = %g"%self.dt)

    def create_scheme(self):
        s = TVFScheme(
            ['fluid'], ['channel'], dim=2, rho0=self.rho0, c0=None,
            nu=self.nu, p0=None, pb=None, h0=h0, gx=None
        )
        return s

    def create_domain(self):
        return DomainManager(xmin=0, xmax=self.Lx, periodic_in_x=True)

    def create_particles(self):
        Lx = self.Lx
        Ly = self.Ly
        _x = np.arange( dx/2, Lx, dx )

        # create the fluid particles
        _y = np.arange( dx/2, Ly, dx )

        x, y = np.meshgrid(_x, _y); fx = x.ravel(); fy = y.ravel()

        # create the channel particles at the top
        _y = np.arange(Ly+dx/2, Ly+dx/2+ghost_extent, dx)
        x, y = np.meshgrid(_x, _y); tx = x.ravel(); ty = y.ravel()

        # create the channel particles at the bottom
        _y = np.arange(-dx/2, -dx/2-ghost_extent, -dx)
        x, y = np.meshgrid(_x, _y); bx = x.ravel(); by = y.ravel()

        # concatenate the top and bottom arrays
        cx = np.concatenate( (tx, bx) )
        cy = np.concatenate( (ty, by) )

        # create the arrays
        channel = get_particle_array(name='channel', x=cx, y=cy)
        fluid = get_particle_array(name='fluid', x=fx, y=fy)

        print("Poiseuille flow :: Re = %g, nfluid = %d, nchannel=%d"%(
            self.re, fluid.get_number_of_particles(),
            channel.get_number_of_particles()))

        # add requisite properties to the arrays:
        self.scheme.setup_properties([fluid, channel])

        # setup the particle properties
        volume = dx * dx

        # mass is set to get the reference density of rho0
        fluid.m[:] = volume * self.rho0
        channel.m[:] = volume * self.rho0

        # Set the default rho.
        fluid.rho[:] = self.rho0
        channel.rho[:] = self.rho0

        # volume is set as dx^2
        fluid.V[:] = 1./volume
        channel.V[:] = 1./volume

        # smoothing lengths
        fluid.h[:] = hdx * dx
        channel.h[:] = hdx * dx

        # return the particle list
        return [fluid, channel]

    def create_tools(self):
        tools = []
        if self.options.remesh > 0:
            from pysph.solver.tools import SimpleRemesher
            remesher = SimpleRemesher(
                self, 'fluid', props=['u', 'v', 'uhat', 'vhat'],
                freq=self.options.remesh
            )
            tools.append(remesher)
        return tools

    def post_process(self, info_fname):
        info = self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')

        y_ex, u_ex, y, u = self._plot_u_vs_y()
        t, ke = self._plot_ke_history()
        res = os.path.join(self.output_dir, "results.npz")
        np.savez(res, t=t, ke=ke, y=y, u=u, y_ex=y_ex, u_ex=u_ex)

    def _plot_ke_history(self):
        from pysph.tools.pprocess import get_ke_history
        from matplotlib import pyplot as plt
        t, ke = get_ke_history(self.output_files, 'fluid')
        plt.clf()
        plt.plot(t, ke)
        plt.xlabel('t')
        plt.ylabel('Kinetic energy')
        fig = os.path.join(self.output_dir, "ke_history.png")
        plt.savefig(fig, dpi=300)
        return t, ke

    def _plot_u_vs_y(self):
        files = self.output_files

        # take the last solution data
        fname = files[-1]
        data = load(fname)
        tf = data['solver_data']['t']
        fluid = data['arrays']['fluid']
        u = fluid.u.copy()
        y = fluid.y.copy()

        # exact parabolic profile for the u-velocity
        d = self.d
        fx = self.fx
        nu = self.nu

        ye = np.linspace(0, self.Ly, 100)
        ue = 0.5 * fx/nu * ye*(self.Ly - ye)
        from matplotlib import pyplot as plt
        plt.clf()
        plt.plot(ye, ue, label="exact")
        plt.plot(y, u, 'ko', fillstyle='none', label="computed")
        plt.xlabel('y')
        plt.ylabel('u')
        plt.legend()
        plt.title('Velocity profile at %s'%tf)
        fig = os.path.join(self.output_dir, "comparison.png")
        plt.savefig(fig, dpi=300)
        return ye, ue, y, u


if __name__ == '__main__':
    app = PoiseuilleFlow()
    app.run()
    app.post_process(app.info_filename)



"""Test if the mpi_reduce_array function works correctly.
"""

import mpi4py.MPI as mpi
import numpy as np

from pysph.base.reduce_array import serial_reduce_array, mpi_reduce_array


def main():
    comm = mpi.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    n = 5
    data = np.ones(n)*(rank + 1)

    full_data = []
    for i in range(size):
        full_data = np.concatenate([full_data, np.ones(n)*(i+1)])

    for op in ('sum', 'prod', 'min', 'max'):
        serial_data = serial_reduce_array(data, op)
        result = mpi_reduce_array(serial_data, op)
        expect = getattr(np, op)(full_data)
        msg = "For op %s: Expected %s, got %s" % (op, expect, result)
        assert expect == result, msg


if __name__ == '__main__':
    main()



# cython: language_level=3, embedsignature=True
# distutils: language=c++

"""Routines for eigen decomposition of symmetric 3x3 matrices.
"""
cdef double det(double a[3][3]) nogil
cdef void get_eigenvalues(double a[3][3], double *result) nogil
cdef void eigen_decomposition(double A[3][3], double V[3][3], double *d) nogil
cdef void transform(double A[3][3], double P[3][3], double res[3][3]) nogil
cdef void transform_diag(double *A, double P[3][3], double res[3][3]) nogil
cdef void transform_diag_inv(double *A, double P[3][3], double res[3][3]) nogil

cdef void get_eigenvalvec(double A[3][3], double *R, double *e)



"""Still water over a submerged hump. (6 sec)

The case is described in "Smoothed Particle Hydrodynamics: Approximate
zero-consistent 2-D boundary conditions and still shallow-water tests",
R. Vacondio, B. D. Rogers and P. K. Stansby, Int. J. Numer. Meth. Fluids,
Vol 69, pp 226-253 (2012). DOI: 10.1002/fld.2559

"""
import os

# Numpy
from numpy import (ones_like, zeros_like, array, pi, savez, arange, sqrt, sin,
                   where)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration, FluidBottomElevation, FluidBottomGradient,
        FluidBottomCurvature, GradientCorrectionPreStep, GradientCorrection,
        BedGradient, BedCurvature
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator


# Constants
rho_w = 1000.0
g = 9.81
dim = 1


class StillWaterOverHump(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.5,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.005,
            help="Spacing between the particles")
        group.add_argument(
            "--fluid_surf_height", action="store", type=float,
            dest="fluid_surf_hei", default=0.4, help="Surface height of fluid\
            column (m)")
        group.add_argument(
            "--le", action="store", type=float, dest="le", default=1.0,
            help="Initial length of the fluid column (m)")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.fluid_surf_hei = self.options.fluid_surf_hei
        self.le = self.options.le

    def create_particles(self):
        # 1D
        hdx = self.hdx
        dx = self.dx
        fluid_surf_hei = self.fluid_surf_hei
        le = self.le

        # Bed
        dxb = dx
        xb = arange(0, le+1e-4, dxb)
        cond = (0.25 < xb) & (xb < 0.75)
        b = where(cond, 0.05*(1+sin(pi*(4*xb+0.5))), 0)

        Vb = ones_like(xb) * dxb
        hb = ones_like(xb) * hdx * dxb

        bed = gpa_swe(name='bed', x=xb, V=Vb, b=b, h=hb)

        # For gradient correction
        len_b = len(bed.x) * 9
        bed.add_constant('m_mat', [0.0] * len_b)

        # Fluid
        x = arange(0, le+1e-4, dx)
        h = ones_like(x) * hdx * dx
        h0 = ones_like(x) * hdx * dx

        fluid = gpa_swe(x=x, h=h, h0=h0, name='fluid')
        compute_fluid_elevation([fluid, bed])

        dw = fluid_surf_hei - fluid.b
        rho = dw * rho_w
        rho0 = dw * rho_w
        m = rho * dx

        fluid.m = m
        fluid.rho = rho
        fluid.rho0 = rho0
        fluid.dw = dw

        compute_initial_props([fluid])

        # Boundary
        x_cb = array([-2*dx, -dx, 1+dx, 1+2*dx])
        m_cb = ones_like(x_cb) * rho_w * dx * fluid_surf_hei
        h_cb = ones_like(x_cb) * hdx * dx
        rho_cb = ones_like(x_cb) * rho_w * fluid_surf_hei
        dw_cb = ones_like(x_cb) * fluid_surf_hei
        cs_cb = sqrt(9.8 * dw_cb)
        alpha_cb = dim * rho_cb
        no_art_visc = ones_like(x_cb, dtype=int)

        boundary = gpa_swe(name='boundary', x=x_cb, m=m_cb,
                           h=h_cb, rho=rho_cb, dw=dw_cb, cs=cs_cb,
                           alpha=alpha_cb, no_art_visc=no_art_visc)

        return [fluid, bed, boundary]

    def create_solver(self):
        kernel = CubicSpline(dim=1)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = 5
        solver = Solver(
            kernel=kernel,
            dim=1,
            integrator=integrator,
            cfl=0.3,
            adaptive_timestep=True,
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(
                                dest='fluid', sources=['fluid', 'boundary']
                                ),
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid', 'boundary']),
                    ]
                ),
            Group(
                equations=[
                    SWEOS(dest='fluid'),
                    ]
                ),
            Group(
                equations=[
                    FluidBottomElevation(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomGradient(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomCurvature(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(
                        dim=dim, dest='fluid', sources=['fluid', 'boundary'],
                        u_only=True),
                    ],
                ),
            ]
        return equations

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth_and_vel()

    def _plot_depth_and_vel(self):
        from matplotlib import pyplot as plt
        t_to_plot = [5.0]
        fname_for_plot = []
        for fname in self.output_files:
            data = load(fname)
            t = data['solver_data']['t']
            if t in t_to_plot:
                fname_for_plot.append(fname)

        num = 0
        for fname in fname_for_plot:
            data = load(fname)
            fluid = data['arrays']['fluid']
            t = data['solver_data']['t']

            x = fluid.x
            x_max = max(x)
            dw = fluid.dw
            dw_max = max(dw)
            u = fluid.u
            u_max = max(u)
            b = fluid.b
            numerical_fluid_surf_hei = dw + b
            exact_fluid_surf_hei = ones_like(x) * self.fluid_surf_hei

            plt.clf()

            plt.plot(x, numerical_fluid_surf_hei, 'b', linewidth=2.0,
                     label='SWE-SPH')
            plt.plot(x, exact_fluid_surf_hei, 'r', linewidth=2.0,
                     label='Exact Solution')
            plt.plot(x, b, 'k--', label='Bottom Elevation')
            plt.ylim(-0.1*self.fluid_surf_hei, 1.2*dw_max)
            plt.xlim(0, x_max)

            plt.legend(prop={'size': 13})
            plt.xlabel('x (m)', fontsize=13)
            plt.ylabel('dw (m)', fontsize=13)
            plt.tick_params(axis='x', labelsize=13)
            plt.tick_params(axis='y', labelsize=13)
            plt.title('t = %0.1f s' % t_to_plot[num], fontsize=13)
            fig = os.path.join(self.output_dir, "depth%d.png" % (num+1))
            plt.savefig(fig, dpi=300)

            plt.clf()

            exact_vel = zeros_like(x)

            fname_res = os.path.join(
                    self.output_dir, 'results.npz')
            savez(
                fname_res, t=t_to_plot[num], x=x, b=b, exact_vel=exact_vel,
                u=u, numerical_fluid_surf_hei=numerical_fluid_surf_hei,
                exact_fluid_surf_hei=exact_fluid_surf_hei)

            plt.plot(x, u, 'b', label='SWE-SPH')
            plt.plot(x, exact_vel, 'r', label='Exact Solution')
            plt.ylim(-1.5*u_max, 1.5*u_max)
            plt.xlim(0, x_max)
            plt.legend(prop={'size': 13})
            plt.xlabel('x (m)', fontsize=13)
            plt.ylabel('u (m/s)', fontsize=13)
            plt.tick_params(axis='x', labelsize=13)
            plt.tick_params(axis='y', labelsize=13)
            plt.title('t = %0.1f s' % t_to_plot[num], fontsize=13)
            plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            fig = os.path.join(self.output_dir, "velocity%d.png" % (num+1))
            plt.savefig(fig, dpi=300)
            num += 1


def compute_fluid_elevation(particles):
    one_time_equations = [
       Group(
            equations=[
                FluidBottomElevation(dest='fluid', sources=['bed'])
                ]
            ),
       Group(
            equations=[
                GradientCorrectionPreStep(dest='bed', sources=['bed']),
                ]
            ),
       Group(
            equations=[
                GradientCorrection(dest='bed', sources=['bed']),
                ]
            ),
       Group(
            equations=[
                BedGradient(dest='bed', sources=['bed']),
                ]
            ),
       Group(
            equations=[
                BedCurvature(dest='bed', sources=['bed']),
                ]
            ),
    ]
    kernel = CubicSpline(dim=1)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=1,
                            kernel=kernel)
    sph_eval.evaluate()


def compute_initial_props(particles):
    one_time_equations = [
        Group(
            equations=[
                SWEOS(dest='fluid')
                    ]
            ),
    ]
    kernel = CubicSpline(dim=1)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=1,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = StillWaterOverHump()
    app.run()
    app.post_process(app.info_filename)



""" Module to run the example files and report their success/failure results

Add a function to the ExampleTest class corresponding to an example script to
be tested.
This is done till better strategy for parallel testing is implemented

"""

from pytest import mark
from .example_test_case import ExampleTestCase, get_example_script
from pysph.base.nnps import get_number_of_threads


@mark.skipif(get_number_of_threads() == 1, reason= "N_threads=1; OpenMP does not seem available.")
class TestOpenMPExamples(ExampleTestCase):

    @mark.slow
    def test_3Ddam_break_example(self):
        dt = 2e-5; tf = 13*dt
        serial_kwargs = dict(
            timestep=dt, tf=tf, pfreq=100, test=None
        )
        extra_parallel_kwargs = dict(openmp=None)
        # Note that we set nprocs=1 here since we do not want
        # to run this with mpirun.
        self.run_example(
            get_example_script('sphysics/dambreak_sphysics.py'),
            nprocs=1, atol=1e-14,
            serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )

    @mark.slow
    def test_elliptical_drop_example(self):
        tf = 0.0076*0.25
        serial_kwargs = dict(kernel='CubicSpline', tf=tf)
        extra_parallel_kwargs = dict(openmp=None)
        # Note that we set nprocs=1 here since we do not want
        # to run this with mpirun.
        self.run_example(
            'elliptical_drop.py', nprocs=1, atol=1e-14,
            serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )

    def test_ldcavity_example(self):
        dt=1e-4; tf=200*dt
        serial_kwargs = dict(timestep=dt, tf=tf, pfreq=500)
        extra_parallel_kwargs = dict(openmp=None)
        # Note that we set nprocs=1 here since we do not want
        # to run this with mpirun.
        self.run_example(
            'cavity.py', nprocs=1, atol=1e-14,
            serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )



from pysph.sph.equation import Equation


class Continuity(Equation):

    def __init__(self, dest, sources, c0):
        self.c0 = c0

        super(Continuity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_arho):
        d_arho[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_u, d_v, d_w, s_u, s_v, s_w, d_cs, s_cs,
             d_rho, d_arho, s_rho, d_p, s_p, DWIJ, RIJ, XIJ):
        rl = d_rho[d_idx]
        rr = s_rho[s_idx]
        pl = d_p[d_idx]
        pr = s_p[s_idx]
        cl = d_cs[d_idx]
        cr = s_cs[s_idx]
        uxl = d_u[d_idx]
        uyl = d_v[d_idx]
        uzl = d_w[d_idx]
        uxr = s_u[s_idx]
        uyr = s_v[s_idx]
        uzr = s_w[s_idx]
        co = self.c0
        eij = declare('matrix(3)')
        vij = declare('matrix(3)')
        v_star = declare('matrix(3)')
        vij[0] = 0.5 * (uxl + uxr)
        vij[1] = 0.5 * (uyl + uyr)
        vij[2] = 0.5 * (uzl + uzr)

        for i in range(3):
            if RIJ >= 1.0e-12:
                eij[i] = -XIJ[i] / RIJ
            else:
                eij[i] = 0.0
        ul = uxl * eij[0] + uyl * eij[1] + uzl * eij[2]
        ur = uxr * eij[0] + uyr * eij[1] + uzr * eij[2]
        rhobar = 0.5 * (rl + rr)
        u_star = 0.5 * (ul + ur) + 0.5 * (pl - pr) / (rhobar * co)
        p_star = 0.5 * (pl + pr) + 0.5 * rhobar * co * (ul - ur)
        for i in range(3):
            v_star[i] = (u_star - 0.5 * (ul + ur)) * eij[i] + vij[i]
        vdotw = (uxl - v_star[0]) * DWIJ[0] + (uyl - v_star[1]) * DWIJ[1]
        vdotw += (uzl - v_star[2]) * DWIJ[2]
        d_arho[d_idx] += 2.0 * s_m[s_idx] * vdotw * rl / rr


class MomentumFluid(Equation):

    def __init__(self, dest, sources, c0, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.c0 = c0

        super(MomentumFluid, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz

    def loop(self, d_idx, s_idx, s_m, d_u, d_v, d_w, s_u, s_v, s_w, d_cs, s_cs,
             d_rho, s_rho, d_p, s_p, d_au, d_av, d_aw, RIJ, XIJ, DWIJ, HIJ):
        rl = d_rho[d_idx]
        rr = s_rho[s_idx]
        pl = d_p[d_idx]
        pr = s_p[s_idx]
        cl = d_cs[d_idx]
        cr = s_cs[s_idx]
        uxl = d_u[d_idx]
        uyl = d_v[d_idx]
        uzl = d_w[d_idx]
        uxr = s_u[s_idx]
        uyr = s_v[s_idx]
        uzr = s_w[s_idx]
        m = s_m[s_idx]
        co = self.c0
        eij = declare('matrix(3)')
        vij = declare('matrix(3)')
        v_star = declare('matrix(3)')
        vij[0] = 0.5 * (uxl + uxr)
        vij[1] = 0.5 * (uyl + uyr)
        vij[2] = 0.5 * (uzl + uzr)

        for i in range(3):
            if RIJ >= 1.0e-12:
                eij[i] = -XIJ[i] / RIJ
            else:
                eij[i] = 0.0
        ul = uxl * eij[0] + uyl * eij[1] + uzl * eij[2]
        ur = uxr * eij[0] + uyr * eij[1] + uzr * eij[2]
        rhobar = 0.5 * (rl + rr)
        p_star = 0.5 * (pl + pr) + 0.5 * rhobar * co * (ul - ur)
        factor = -2.0 * m * p_star / (rl * rr)

        d_au[d_idx] += factor * DWIJ[0]
        d_av[d_idx] += factor * DWIJ[1]
        d_aw[d_idx] += factor * DWIJ[2]



import numpy as np
import unittest
from pytest import importorskip

cl = importorskip('pyopencl')

import pysph.base.particle_array
from pysph.base.device_helper import DeviceHelper   # noqa: E402
from pysph.base.utils import get_particle_array  # noqa: E402
from pysph.base.tree.point_tree import PointTree  # noqa: E402


def _gen_uniform_dataset_2d(n, h, seed=None):
    if seed is not None:
        np.random.seed(seed)
    u = np.random.uniform
    pa = get_particle_array(x=u(size=n), y=u(size=n), h=h)
    h = DeviceHelper(pa, backend='opencl')
    pa.set_device_helper(h)

    return pa


def _gen_uniform_dataset(n, h, seed=None):
    if seed is not None:
        np.random.seed(seed)
    u = np.random.uniform
    pa = get_particle_array(x=u(size=n), y=u(size=n), z=u(size=n), h=h)
    h = DeviceHelper(pa, backend='opencl')
    pa.set_device_helper(h)

    return pa


def _dfs_find_leaf(tree):
    leaf_id_count = tree.allocate_leaf_prop(np.int32)
    dfs_find_leaf = tree.leaf_tree_traverse(
        "int *leaf_id_count",
        setup="leaf_id_count[i] = 0;",
        node_operation="if (cid_dst == cid_src) leaf_id_count[i]++",
        leaf_operation="if (cid_dst == cid_src) leaf_id_count[i]++",
        output_expr=""
    )

    dfs_find_leaf(tree, tree, leaf_id_count.dev)
    return leaf_id_count.dev.get()


def _check_children_overlap_2d(node_xmin, node_xmax, child_offset):
    for j in range(4):
        nxmin1 = node_xmin[child_offset + j]
        nxmax1 = node_xmax[child_offset + j]
        for k in range(4):
            nxmin2 = node_xmin[child_offset + k]
            nxmax2 = node_xmax[child_offset + k]
            if j != k:
                assert (nxmax1[0] <= nxmin2[0] or nxmax2[0] <= nxmin1[0] or
                        nxmax1[1] <= nxmin2[1] or nxmax2[1] <= nxmin1[1])


def _check_children_overlap(node_xmin, node_xmax, child_offset):
    for j in range(8):
        nxmin1 = node_xmin[child_offset + j]
        nxmax1 = node_xmax[child_offset + j]
        for k in range(8):
            nxmin2 = node_xmin[child_offset + k]
            nxmax2 = node_xmax[child_offset + k]
            if j != k:
                assert (nxmax1[0] <= nxmin2[0] or nxmax2[0] <= nxmin1[0] or
                        nxmax1[1] <= nxmin2[1] or nxmax2[1] <= nxmin1[1] or
                        nxmax1[2] <= nxmin2[2] or nxmax2[2] <= nxmin1[2])


def _test_tree_structure(tree, k):
    # Traverse tree and check if max depth is correct
    # Additionally check if particle sets of siblings is disjoint
    # and union of particle sets of a nodes children = nodes own children
    #
    # This effectively also checks that no  particle is present in two nodes of
    # the same level

    s = [0, ]
    d = [0, ]

    offsets = tree.offsets.dev.get()
    pbounds = tree.pbounds.dev.get()

    max_depth = tree.depth
    max_depth_here = 0
    pids = set()

    while len(s) != 0:
        n = s[0]
        depth = d[0]
        max_depth_here = max(max_depth_here, depth)
        pbound = pbounds[n]
        assert (depth <= max_depth)

        del s[0]
        del d[0]

        if offsets[n] == -1:
            for i in range(pbound[0], pbound[1]):
                pids.add(i)
            continue

        # Particle ranges of children are contiguous
        # and are contained within parent's particle range
        start = pbound[0]
        for i in range(k):
            child_idx = offsets[n] + i
            assert (pbounds[child_idx][0] == start)
            assert (pbounds[child_idx][0] <= pbounds[child_idx][1])
            start = pbounds[child_idx][1]

            assert (child_idx < len(offsets))
            s.append(child_idx)
            d.append(depth + 1)
        assert (start == pbound[1])


class QuadtreeTestCase(unittest.TestCase):
    def setUp(self):
        use_double = False
        self.N = 3000
        pa = _gen_uniform_dataset_2d(self.N, 0.2, seed=0)
        self.quadtree = PointTree(pa, radius_scale=1., use_double=use_double,
                                  leaf_size=32, dim=2)
        self.leaf_size = 32
        self.quadtree.refresh(np.array([0., 0.]), np.array([1., 1.]),
                              np.min(pa.h))
        self.pa = pa

    def test_pids(self):
        pids = self.quadtree.pids.dev.get()
        s = set()
        for i in range(len(pids)):
            if 0 <= pids[i] < self.N:
                s.add(pids[i])

        assert (len(s) == self.N)

    def test_depth_and_inclusiveness(self):
        _test_tree_structure(self.quadtree, 4)

    def test_node_bounds(self):

        self.quadtree.set_node_bounds()
        pids = self.quadtree.pids.dev.get()
        offsets = self.quadtree.offsets.dev.get()
        pbounds = self.quadtree.pbounds.dev.get()
        node_xmin = self.quadtree.node_xmin.dev.get()
        node_xmax = self.quadtree.node_xmax.dev.get()
        node_hmax = self.quadtree.node_hmax.dev.get()

        x = self.pa.x[pids]
        y = self.pa.y[pids]
        h = self.pa.h[pids]

        for i in range(len(offsets)):
            nxmin = node_xmin[i]
            nxmax = node_xmax[i]
            nhmax = node_hmax[i]

            for j in range(pbounds[i][0], pbounds[i][1]):
                assert (nxmin[0] <= np.float32(x[j]) <= nxmax[0])
                assert (nxmin[1] <= np.float32(y[j]) <= nxmax[1])
                assert (np.float32(h[j]) <= nhmax)
            # Check that children nodes don't overlap
            if offsets[i] != -1:
                _check_children_overlap_2d(node_xmin, node_xmax, offsets[i])

    def test_dfs_traversal(self):
        leaf_id_count = _dfs_find_leaf(self.quadtree)
        np.testing.assert_array_equal(
            np.ones(self.quadtree.unique_cid_count, dtype=np.int32),
            leaf_id_count
        )

    def test_get_leaf_size_partitions(self):
        a, b = np.random.randint(0, self.leaf_size, size=2)
        a, b = min(a, b), max(a, b)

        pbounds = self.quadtree.pbounds.dev.get()
        offsets = self.quadtree.offsets.dev.get()

        mapping, count = self.quadtree.get_leaf_size_partitions(a, b)
        mapping = mapping.dev.get()
        map_set_gpu = {mapping[i] for i in range(count)}
        map_set_here = {i for i in range(len(offsets))
                        if offsets[i] == -1 and
                        a < (pbounds[i][1] - pbounds[i][0]) <= b}
        assert (map_set_gpu == map_set_here)

    def tearDown(self):
        del self.quadtree


class OctreeTestCase(unittest.TestCase):
    def setUp(self):
        use_double = False
        self.N = 3000
        pa = _gen_uniform_dataset(self.N, 0.2, seed=0)
        self.octree = PointTree(pa, dim=3, radius_scale=1.,
                                use_double=use_double,
                                leaf_size=64)
        self.leaf_size = 64
        self.octree.refresh(np.array([0., 0., 0.]), np.array([1., 1., 1.]),
                            np.min(pa.h))
        self.pa = pa

    def test_pids(self):
        pids = self.octree.pids.dev.get()
        s = set()
        for i in range(len(pids)):
            if 0 <= pids[i] < self.N:
                s.add(pids[i])

        assert (len(s) == self.N)

    def test_depth_and_inclusiveness(self):
        _test_tree_structure(self.octree, 8)

    def test_node_bounds(self):

        self.octree.set_node_bounds()
        print(self.octree.node_hmax.dev.get())
        pids = self.octree.pids.dev.get()
        offsets = self.octree.offsets.dev.get()
        pbounds = self.octree.pbounds.dev.get()
        node_xmin = self.octree.node_xmin.dev.get()
        node_xmax = self.octree.node_xmax.dev.get()
        node_hmax = self.octree.node_hmax.dev.get()

        x = self.pa.x[pids]
        y = self.pa.y[pids]
        z = self.pa.z[pids]
        h = self.pa.h[pids]

        for i in range(len(offsets)):
            nxmin = node_xmin[i]
            nxmax = node_xmax[i]
            nhmax = node_hmax[i]

            for j in range(pbounds[i][0], pbounds[i][1]):
                assert (nxmin[0] <= np.float32(x[j]) <= nxmax[0])
                assert (nxmin[1] <= np.float32(y[j]) <= nxmax[1])
                assert (nxmin[2] <= np.float32(z[j]) <= nxmax[2])
                assert (np.float32(h[j]) <= nhmax)
            # Check that children nodes don't overlap
            if offsets[i] != -1:
                _check_children_overlap(node_xmin, node_xmax, offsets[i])

    def test_dfs_traversal(self):
        leaf_id_count = _dfs_find_leaf(self.octree)
        np.testing.assert_array_equal(
            np.ones(self.octree.unique_cid_count, dtype=np.int32),
            leaf_id_count
        )

    def test_get_leaf_size_partitions(self):
        a, b = np.random.randint(0, self.leaf_size, size=2)
        a, b = min(a, b), max(a, b)

        pbounds = self.octree.pbounds.dev.get()
        offsets = self.octree.offsets.dev.get()

        mapping, count = self.octree.get_leaf_size_partitions(a, b)
        mapping = mapping.dev.get()
        map_set_gpu = {mapping[i] for i in range(count)}
        map_set_here = {i for i in range(len(offsets))
                        if offsets[i] == -1 and
                        a < (pbounds[i][1] - pbounds[i][0]) <= b}
        assert (map_set_gpu == map_set_here)

    def tearDown(self):
        del self.octree



"""Rectangular open channel flow over an inclined bed with friction. (4 min)

The case is described in "Shallow Water and Navier-Stokes SPH-like numerical
modelling of rapidly varying free-surface flows", Renato Vacondio, Doctoral
dissertation, University of Parma, pp 99-103 (2010).

"""
import os

# Numpy
from numpy import (ones_like, zeros_like, sort, arange, sqrt, savez)
import numpy as np

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration, BoundaryInnerReimannStateEval, SubCriticalInFlow,
        SubCriticalOutFlow, BedFrictionSourceEval
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator

# PySPH Inlet_Outlet
from pysph.sph.bc.donothing.simple_inlet_outlet import (
    SimpleInletOutlet)
from pysph.sph.bc.inlet_outlet_manager import (
    InletInfo, OutletInfo, OutletStep, InletStep)

# Constants
rho_w = 1000.0
g = 9.81
dim = 2


class RectangularOpenChannelFlow(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.2,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=10.0,
            help="Spacing between the particles")
        group.add_argument(
            "--dw0", action="store", type=float, dest="dw0", default=5.0,
            help="Initial depth of the fluid column (m)")
        group.add_argument(
            "--le", action="store", type=float, dest="le", default=800.0,
            help="Length of the channel (m)")
        group.add_argument(
            "--w", action="store", type=float, dest="w", default=400.0,
            help="Width of the channel (m)")
        group.add_argument(
            "--n", action="store", type=float, dest="n", default=0.0316,
            help="Manning coefficient ")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.dw0 = self.options.dw0
        self.le = self.options.le
        self.w = self.options.w
        self.n = self.options.n

        # Inlet and Outlet
        self.num_inlet_pa = 2
        self.num_outlet_pa = 3
        self.x_max_inlet = 0.
        self.x_min_inlet = -self.dx * self.num_inlet_pa
        self.x_min_outlet = self.le
        self.x_max_outlet = self.le + self.num_outlet_pa*self.dx

        # Initial Condition
        q = 14.645  # Specific Discharge
        self.u_inlet = q / self.dw0

    def create_particles(self):
        hdx = self.hdx
        dx = self.dx
        d = self.dw0
        w = self.w
        le = self.le
        # Inlet Properties
        x, y = np.mgrid[-self.num_inlet_pa*dx + dx/2.:0:dx, dx/2:w-(dx/4.):dx]
        x, y = (np.ravel(t) for t in (x, y))

        u_inlet = self.u_inlet

        m = ones_like(x) * dx * dx * rho_w * d
        h = ones_like(x) * hdx * dx
        h0 = ones_like(x) * hdx * dx

        rho = ones_like(x) * rho_w * d
        rho0 = ones_like(x) * rho_w * d
        alpha = dim * rho
        cs = sqrt(9.8 * rho/rho_w)

        # Note: Both u and uh must be specified when specifying inlet vel
        u = ones_like(x) * u_inlet
        uh = ones_like(x) * u_inlet

        # Need to specify this to inlet open boundary particle (OBP) because
        # when OBP becomes fluid particle, the fluid particle will have this
        # bottom slope (bx)
        bx = ones_like(x) * -0.001

        inlet = gpa_swe(x=x, y=y, m=m, rho0=rho0, rho=rho, h0=h0, h=h, u=u,
                        uh=uh, alpha=alpha, cs=cs, bx=bx, name='inlet')
        boundary_props = ['dw_inner_reimann', 'u_inner_reimann',
                          'v_inner_reimann', 'shep_corr']
        inlet.add_output_arrays(boundary_props)
        inlet.add_property('x0')

        # Fluid Properties
        xf, yf = np.mgrid[0.5*dx:self.x_max_inlet+le:dx, dx/2:w-(dx/4.):dx]
        xf, yf = (np.ravel(t) for t in (xf, yf))
        m = ones_like(xf) * dx * dx * rho_w * d
        h = ones_like(xf) * hdx * dx
        h0 = ones_like(xf) * hdx * dx
        rho = ones_like(xf) * rho_w * d
        rho0 = ones_like(xf) * rho_w * d
        uh = ones_like(xf) * u_inlet
        u = ones_like(xf) * u_inlet
        bx = ones_like(xf) * -0.001
        fluid = gpa_swe(name='fluid', x=xf, y=yf, m=m, rho0=rho0, rho=rho, h=h,
                        bx=bx, h0=h0, uh=uh, u=u)

        # Outlet Properties
        xo, yo = np.mgrid[dx/2.:self.num_outlet_pa*dx:dx, dx/2:w-(dx/4.):dx]
        xo, yo = (np.ravel(t) for t in (xo, yo))
        xo += le
        dw = ones_like(xo) * d
        m = ones_like(xo) * dx * dx * rho_w * d
        h = ones_like(xo) * hdx * dx
        h0 = ones_like(xo) * hdx * dx
        rho = ones_like(xo) * rho_w * d
        rho0 = ones_like(xo) * rho_w * d
        cs = sqrt(9.8 * rho/rho_w)
        alpha = dim * rho
        outlet = gpa_swe(name='outlet', x=xo, y=yo, dw=dw, m=m, rho0=rho0,
                         alpha=alpha, rho=rho, h=h, h0=h0, cs=cs)
        outlet.add_output_arrays(boundary_props)
        outlet.add_property('x0')

        # Bed Properties
        xb, yb = np.mgrid[-5*dx:le*1.6+5*dx:dx, 0:w+dx/2.:dx]
        xb = np.ravel(xb)
        yb = np.ravel(yb)

        Vb = ones_like(xb) * dx * dx
        nb = ones_like(xb) * self.n  # Manning Coefficient
        hb = ones_like(xb) * hdx * dx

        bed = gpa_swe(name='bed', x=xb, y=yb, V=Vb, n=nb, h=hb)

        # Closed Boundary
        xcb_top = np.arange(self.x_min_inlet-2.0*dx,
                            self.x_max_outlet*1.6, dx)
        ycb_top = np.concatenate((ones_like(xcb_top)*(w+0.5*dx),
                                  ones_like(xcb_top)*(w+1.5*dx)), axis=0)
        xcb_top = np.tile(xcb_top, 2)

        xcb_bottom = np.arange(self.x_min_inlet-2.0*dx,
                               self.x_max_outlet*1.6, dx)
        ycb_bottom = np.concatenate((zeros_like(xcb_bottom)-0.5*dx,
                                     zeros_like(xcb_bottom)-1.5*dx), axis=0)
        xcb_bottom = np.tile(xcb_bottom, 2)

        xcb_all = np.concatenate((xcb_top, xcb_bottom), axis=0)
        ycb_all = np.concatenate((ycb_top, ycb_bottom), axis=0)
        m_cb = ones_like(xcb_all) * dx * dx * rho_w * d
        h_cb = ones_like(xcb_all) * hdx * dx
        rho_cb = ones_like(xcb_all) * rho_w * d
        dw_cb = ones_like(xcb_all) * d
        cs_cb = sqrt(9.8 * dw_cb)
        alpha_cb = dim * rho_cb
        u_cb = ones_like(xcb_all) * u_inlet
        is_wall_boun_pa = ones_like(xcb_all)

        boundary = gpa_swe(name='boundary', x=xcb_all, y=ycb_all, m=m_cb,
                           h=h_cb, rho=rho_cb, dw=dw_cb, cs=cs_cb,
                           alpha=alpha_cb, u=u_cb,
                           is_wall_boun_pa=is_wall_boun_pa)
        return [inlet, fluid, outlet, bed, boundary]

    def _create_inlet_outlet_manager(self):
        from pysph.sph.bc.donothing.inlet import Inlet
        from pysph.sph.bc.donothing.outlet import Outlet

        props_to_copy = ['x', 'y', 'u', 'v', 'w', 'm',
                         'h', 'rho', 'rho0', 'bx', 'h0', 'uh']
        inlet_info = InletInfo(
            pa_name='inlet', normal=[-1.0, 0.0, 0.0],
            refpoint=[self.x_min_inlet, 0.0, 0.0], has_ghost=False,
            update_cls=Inlet
        )

        outlet_info = OutletInfo(
            pa_name='outlet', normal=[1.0, 0.0, 0.0],
            refpoint=[self.x_max_outlet, 0.0, 0.0], update_cls=Outlet,
            props_to_copy=props_to_copy
        )

        iom = SimpleInletOutlet(
            fluid_arrays=['fluid'], inletinfo=[inlet_info],
            outletinfo=[outlet_info]
        )

        return iom

    def create_inlet_outlet(self, particle_arrays):
        iom = self.iom
        compute_initial_props(list(particle_arrays.values()))
        io = iom.get_inlet_outlet(particle_arrays)
        return io

    def create_solver(self):
        self.iom = self._create_inlet_outlet_manager()
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(inlet=InletStep(), fluid=SWEStep(),
                                   outlet=OutletStep())
        tf = 100
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.1,
            adaptive_timestep=True,
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    GatherDensityEvalNextIteration(
                        dest='fluid', sources=['inlet', 'fluid',
                                               'outlet', 'boundary']),
                    NonDimensionalDensityResidual(dest='fluid'),
                    UpdateSmoothingLength(dim=dim, dest='fluid'),
                    CheckConvergenceDensityResidual(dest='fluid')
                ], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid', 'inlet', 'outlet',
                                               'boundary']),
                    SWEOS(dest='fluid')
                ]
            ),
            Group(
                equations=[
                    BoundaryInnerReimannStateEval(dest='inlet',
                                                  sources=['fluid']),
                    BoundaryInnerReimannStateEval(dest='outlet',
                                                  sources=['fluid'])
                    ]
                ),
            Group(
                equations=[
                    SubCriticalInFlow(dest='inlet'),
                    SubCriticalOutFlow(dest='outlet')
                    ]
                ),
            Group(
                equations=[
                    BedFrictionSourceEval(
                        dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(dim=dim, dest='fluid',
                                         sources=['fluid', 'inlet', 'outlet',
                                                  'boundary'])
                    ]
                ),
            ]
        return equations

    def post_step(self, solver):
        for pa in self.particles:
            if pa.name == 'outlet':
                o_pa = pa
        arr_ones = ones_like(o_pa.rho)
        o_pa.alpha = arr_ones * dim * rho_w * self.dw0
        o_pa.rho = arr_ones * rho_w * self.dw0
        o_pa.dw = arr_ones * self.dw0
        o_pa.cs = sqrt(9.8 * o_pa.dw)

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_vel_and_depth_at_channel_mid_section()

    def _plot_vel_and_depth_at_channel_mid_section(self):
        from matplotlib import pyplot as plt
        t_to_plot = [100.]
        fname_for_plot = []
        for fname in self.output_files:
            data = load(fname)
            t = data['solver_data']['t']
            if t in t_to_plot:
                fname_for_plot.append(fname)

        num = 0
        for fname in fname_for_plot:
            data = load(fname)
            fluid = data['arrays']['fluid']

            y = arange(self.dx/2., self.w-(self.dx/4.), self.dx)
            y_srt = sort(y)
            len_y = len(y_srt)
            idx_mid_y = int(len_y/2.)
            mid_y = y_srt[idx_mid_y]
            cond = abs(fluid.y - mid_y) < 1e-2

            x = fluid.x[cond]
            y = fluid.y[cond]
            dw = fluid.dw[cond]
            rho = fluid.rho[cond]
            u = fluid.u[cond]
            v = fluid.v[cond]
            au = fluid.au[cond]
            av = fluid.av[cond]

            if len(x) == 0:
                return
            x_ind_sort = x.argsort()
            x = x[x_ind_sort]
            y = y[x_ind_sort]
            x_max = max(x)
            dw = dw[x_ind_sort]
            rho = rho[x_ind_sort]
            u = u[x_ind_sort]
            v = v[x_ind_sort]
            au = au[x_ind_sort]
            av = av[x_ind_sort]

            dw_actual = ones_like(x) * self.dw0
            u_actual = ones_like(x) * self.u_inlet

            fname_res = os.path.join(
                    self.output_dir, 'results.npz')
            savez(
                fname_res, t=t_to_plot[num], x=x, y=y,
                dw=dw, u=u, v=v, rho=rho, au=au, av=av)

            # Plot properties of the fluid domain
            plt.clf()
            plt.plot(x, dw_actual, 'r-', label='Exact Solution')
            plt.plot(x, dw, 'bo', label='SWE-SPH', markersize=2)
            plt.ylim(0.5*self.dw0, 1.5*self.dw0)
            plt.xlim(0, x_max+self.dx)
            plt.legend()
            plt.xlabel('x (m)')
            plt.ylabel('dw (m)')
            plt.title('t = %0.1fs' % t_to_plot[num])
            fig = os.path.join(self.output_dir, "depth%d.png" % (num+1))
            plt.savefig(fig, dpi=300)

            plt.clf()
            plt.plot(x, u_actual, 'r-', label='Exact Solution')
            plt.plot(x, u, 'bo', label='SWE-SPH', markersize=2)
            plt.ylim(0.5*self.u_inlet, 1.5*self.u_inlet)
            plt.xlim(0, x_max+self.dx)
            plt.legend()
            plt.xlabel('x (m)')
            plt.ylabel('u (m/s)')
            plt.title('t = %0.1fs' % t_to_plot[num])
            fig = os.path.join(self.output_dir, "velocity%d.png" % (num+1))
            plt.savefig(fig, dpi=300)
            num += 1


def compute_initial_props(particles):
    one_time_equations = [
                Group(
                    equations=[
                        SWEOS(dest='fluid'),
                        ]
                    ),
                Group(
                    equations=[
                        BoundaryInnerReimannStateEval(dest='inlet',
                                                      sources=['fluid']),
                        BoundaryInnerReimannStateEval(dest='outlet',
                                                      sources=['fluid'])
                        ]
                    ),
                Group(
                    equations=[
                        SubCriticalInFlow(dest='inlet'),
                        SubCriticalOutFlow(dest='outlet')
                        ]
                    ),
                Group(
                    equations=[
                        CorrectionFactorVariableSmoothingLength(
                            dest='fluid', sources=['fluid', 'inlet', 'outlet',
                                                   'boundary'])
                        ]
                    ),
            ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == "__main__":
    app = RectangularOpenChannelFlow()
    app.run()
    app.post_process(app.info_filename)



# See PEP 440 for more on suitable version numbers.
__version__ = '1.0b1'

# Utility functions to determine if Zoltan/MPI are available.
_has_zoltan = None
_has_opencl = None
_has_mpi = None
_in_parallel = None


try:
    from pyzoltan import has_mpi  # noqa: 402
except ImportError:
    def has_mpi():
        global _has_mpi
        if _has_mpi is None:
            try:
                import mpi4py  # noqa: 401
            except ImportError:
                _has_mpi = False
            else:
                mpi4py.rc.initialize = False
                mpi4py.rc.finalize = True
        return _has_mpi


def has_opencl():
    """Return True if pyopencl is available.
    """
    global _has_opencl
    if _has_opencl is None:
        _has_opencl = True
        try:
            import pyopencl  # noqa: 401
        except ImportError:
            _has_opencl = False
    return _has_opencl


def has_zoltan():
    """Return True if zoltan is available.
    """
    global _has_zoltan
    if _has_zoltan is None:
        _has_zoltan = True
        try:
            from pyzoltan.core import zoltan  # noqa: 401
        except ImportError:
            _has_zoltan = False
    return _has_zoltan


def in_parallel():
    """Return true if we're running with MPI and Zoltan support
    """
    global _in_parallel
    if _in_parallel is None:
        _in_parallel = has_mpi() and has_zoltan()

    return _in_parallel


# Utility function to determine the possible output files
_has_h5py = None
_has_pyvisfile = None
_has_tvtk = None


def has_h5py():
    """Return True if h5py is available.
    """
    global _has_h5py
    if _has_h5py is None:
        _has_h5py = True
        try:
            import h5py  # noqa: 401
        except ImportError:
            _has_h5py = False
    return _has_h5py


def has_tvtk():
    """Return True if tvtk is available.
    """
    global _has_tvtk
    if _has_tvtk is None:
        _has_tvtk = True
        try:
            import tvtk  # noqa: 401
        except ImportError:
            _has_tvtk = False
    return _has_tvtk


def has_pyvisfile():
    """Return True if pyvisfile is available.
    """
    global _has_pyvisfile
    if _has_pyvisfile is None:
        _has_pyvisfile = True
        try:
            import pyvisfile  # noqa: 401
        except ImportError:
            _has_pyvisfile = False
    return _has_pyvisfile
